import os
import unittest
import datetime
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.room import Room
from potnanny.models.action import Action, ActionSchema, ActionTrigger
from potnanny.plugins.base import ActionPlugin

CLEANUP = True
DBFILE = '/tmp/test_action.db'

class MyActionPlugin(ActionPlugin):
    def __init__(self):
        pass

    async def input(self, data):
        for k,v in data.items():
            print(f"data | {k}: {v}")


async def init_tables():
    try:
        db.init(f'aiosqlite:///{DBFILE}', pragmas=(('foreign_keys', 1),))
        db.register(Room)
        db.register(Action)
        db.register(ActionTrigger)
        await Room.create_table()
        await Action.create_table()
        await ActionTrigger.create_table()

        r = await Room.create(
            name='test room')
        await r.save()

        d = await Device.create(
            name='test device',
            interface='foo.bar.Baz',
            room_id=r.id)
        await d.save()
    except:
        pass


class TestModels(IsolatedAsyncioTestCase):
    @classmethod
    def tearDownClass(cls):
        if CLEANUP:
            try:
                os.unlink(DBFILE)
            except:
                pass

    async def asyncSetUp(self):
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        obj = await Action.create(
            name='test action',
            room_id=1,
            interface='__main__.MyActionPlugin',
            attributes={
                'condition': '>=',
                'threshold': 90
            }
        )
        await obj.save()

        m = {
            'device_id': 1,
            'value': 99.0,
            'type': 'temperature'}

        await obj.input(m)


if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
import datetime
import marshmallow
import calendar
from peewee_aio import fields
from potnanny.database import BaseModel
from potnanny.models.schemas.safe import SafeSchema
from potnanny.models.weekday import WeekdayMap
from potnanny.controllers.outlet import switch_device_outlet
from .device import Device


logger = logging.getLogger(__name__)


class ScheduleSchema(SafeSchema):
    name = marshmallow.fields.String(allow_none=False)
    on_time = marshmallow.fields.String(allow_none=False)
    off_time = marshmallow.fields.String(allow_none=False)
    days = marshmallow.fields.Integer(allow_none=False)
    outlet = marshmallow.fields.Integer(allow_none=False)
    device_id = marshmallow.fields.Integer(allow_none=True)


class Schedule(BaseModel):
    id = fields.AutoField()
    name = fields.CharField(48)
    days = fields.IntegerField()
    on_time = fields.TimeField()
    off_time = fields.TimeField()
    outlet = fields.IntegerField()
    created = fields.DateTimeField(default=datetime.datetime.utcnow)
    device = fields.ForeignKeyField(Device,
        on_delete='CASCADE',
        backref='schedules' )

    def __str__(self):
        return f"<Schedule id={self.id}>"

    def as_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'days': self.days,
            'on_time': ":".join(str(self.on_time).split(":")[0:2]),
            'off_time': ":".join(str(self.off_time).split(":")[0:2]),
            'outlet': self.outlet,
            'device_id': self.device_id,
            'created': self.created.isoformat() + "Z",
        }

    def runs_now(self,
        now:datetime.datetime = datetime.datetime.now(),
        show_key: bool = True):

        """
        Does this schedule need to run now?
        """

        possible_days = WeekdayMap.weekdays_from_value(self.days)
        dow = calendar.day_name[now.weekday()]
        if dow not in possible_days:
            if show_key:
                return (False, None)
            else:
                return False

        for key in ['on', 'off']:
            label = f"{key}_time"
            if not hasattr(self, label):
                continue

            try:
                hh, mm = str(getattr(self, label)).split(':')[0:2]
                if int(hh) == int(now.hour) and int(mm) == int(now.minute):
                    if show_key is True:
                        return (True, key)
                    else:
                        return True
            except Exception as x:
                logger.warning(x)

        if show_key is True:
            return (False, None)
        else:
            return False


    async def run(self, now:datetime.datetime = datetime.datetime.now()):
        """
        Run this schedule now
        """

        true, key = self.runs_now(now, True)
        if true is not True:
            return

        logger.debug(f"This schedule runs now! {self}")
        try:
            state = 1
            if (key == 'off'):
                state = 0

            logger.debug("Switching device %d, outlet %d %s" % (
                self.device_id, self.outlet, key.upper()))
            rval = await switch_device_outlet(
                self.device_id, self.outlet, state)
            # sleep a while, and switch it again... just to be sure
            await asyncio.sleep(45)
            rval = await switch_device_outlet(
                self.device_id, self.outlet, state)
        except Exception as x:
            logger.warning(x)

        return



from .base import (DevicePlugin, ActionPlugin, PipelinePlugin,
    BluetoothDevicePlugin, GPIODevicePlugin)

__all__ = [DevicePlugin, BluetoothDevicePlugin, GPIODevicePlugin, ActionPlugin,
    PipelinePlugin]



import logging
import datetime
import marshmallow
from peewee_aio import fields
from potnanny.database import BaseModel
from potnanny.models.mixins import DeviceMixin
from potnanny.models.schemas.safe import SafeSchema
from .room import Room
from .ext import JSONField


logger = logging.getLogger(__name__)


class DeviceSchema(SafeSchema):
    name = marshmallow.fields.String()
    notes = marshmallow.fields.String(allow_none=True)
    room_id = marshmallow.fields.Integer(allow_none=True)
    interface = marshmallow.fields.String()
    attributes = marshmallow.fields.Dict(allow_none=True)


class Device(BaseModel, DeviceMixin):
    id = fields.AutoField()
    name = fields.CharField(48)
    interface = fields.CharField(64)
    attributes = JSONField(default={})
    created = fields.DateTimeField(default=datetime.datetime.utcnow)
    room = fields.ForeignKeyField(Room,
        null=True,
        on_delete='CASCADE',
        backref='devices')

    def __str__(self):
        return f"<Device id={self.id}, name='{self.name}'>"

    def as_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'interface': self.interface,
            'attributes': self.attributes,
            'created': self.created.isoformat() + "Z",
            'room_id': self.room_id,
        }

        try:
            mylist = self.plugin.reports
            if ('temperature' in mylist and
                'humidity' in mylist and 'vpd' not in mylist):
                mylist.append('vpd')
            data['reports'] = mylist
        except:
            pass

        try:
            for k, v in self.plugin.__dict__.items():
                if 'plugin' not in data:
                    data['plugin'] = {}

                if k.startswith('_'):
                    continue

                data['plugin'][k] = v
        except:
            pass

        return data



import asyncio
import datetime
import logging
import random
from peewee import fn
from potnanny.database import db
from potnanny.database import lock as db_lock
from potnanny.models.device import Device
from potnanny.models.measurement import Measurement
from potnanny.ble import lock as ble_lock


logger = logging.getLogger(__name__)


class OutletControl:
    """
    Class to interface with device switchable outlets
    """

    def __init__(self, device_id: int, outlet_id: int):
        self.device_id = device_id
        self.outlet_id = outlet_id
        self.retries = 5
        self.device = None


    async def switch(self, state: int, force: bool = False) -> int:
        """
        Switch device to requested state
        returns:
            0   = outlet is now off
            1   = outlet is now on
           -1   = outlet was not changed (maybe error, maybe is already in
                  the requested state)
        """

        if not force:
            latest = await self.last_state()
            if latest == state:
                return state

        success = False
        tries = self.retries

        while success is not True and tries:
            try:
                async with ble_lock:
                    success = await self._switch_outlet(state)

                if success:
                    break
            except Exception as x:
                logger.warning(str(x))
                pass

            tries -= 1
            await asyncio.sleep(random.random())

        if success:
            return state
        else:
            logger.warning(
                f"Switch fail device {self.device_id}, outlet {self.outlet_id}"
            )
            return -1


    async def last_state(self, timeout: int = 600) -> int:
        """
        Get most recent recorded db measurement of outlet state
        return will be the last state (0|1), or -1 if could not be found.
        """

        now = datetime.datetime.utcnow().replace(microsecond=0)
        start = now - datetime.timedelta(seconds=timeout)

        try:
            results = await Measurement.select(
                Measurement.id,
                Measurement.type,
                Measurement.value,
                Measurement.device_id,
                fn.MAX(Measurement.created)
            ).where(
                Measurement.device_id == self.device_id,
                Measurement.type == f"outlet_{self.outlet_id}"
            ).group_by(
                Measurement.device_id
            )
            m = results[0]
            if m.created < start:
                return -1
            return int(m.value)
        except Exception as x:
            logger.warning(str(x))

        return -1


    async def _switch_outlet(self, state):
        if self.device is None:
            self.device = await Device.get_by_id(self.device_id)

        if not hasattr(self.device.plugin, 'is_switchable') or \
            self.device.plugin.is_switchable is not True:
            logger.warning(f"Device with id {device_id} is not switchable")
            return False

        logger.debug("switching device...")
        if bool(state):
            rval = await self.device.on(self.outlet_id)
            logger.debug(f"switch rval: {rval}")
            if int(rval) == 1:
                return True
        else:
            rval = await self.device.off(self.outlet_id)
            logger.debug(f"switch rval: {rval}")
            if int(rval) == 0:
                return True

        return False


async def switch_device_outlet(device_id:int, outlet_id:int, state:int):
    """
    Convenience function for times when we need a basic function
    like... a functools.partial call
    """

    ctl = OutletControl(device_id, outlet_id)
    result = await ctl.switch(state, True)
    if result == state:
        async with db_lock:
            async with db.transaction():
                try:
                    m = await Measurement.create(
                        created=datetime.datetime.utcnow().replace(microsecond=0),
                        type=f"outlet_{outlet_id}",
                        value=state,
                        device_id=device_id
                    )
                except Exception as x:
                    logger.warning(str(x))

    return result



import asyncio
import random
import bleak
import logging
import datetime
import itertools
from potnanny.ble import lock
from potnanny.utils import utcnow
from potnanny.models.device import Device
from potnanny.controllers.pipeline import Pipeline
from .parser import Parser


logger = logging.getLogger(__name__)


class Collector:

    def __init__(self, *args, **kwargs):
        self.convert_c = False
        self.leaf_offset = -2
        self.devices = []
        self.now = utcnow()

        for k,v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)


    async def collect(self):
        """
        Collect measurements from bluetooth devices, and send to pipeline
        """

        measurements = []
        results = {}
        if not self.devices:
            await self.load_devices()

        readers = self.get_reader_list(self.devices)
        clients = self.get_poller_list(self.devices)

        if readers:
            data = await self.scan_advertised(readers)
            if data:
                results.update(data)

        if clients:
            data = await self.poll_clients(clients)
            if data:
                results.update(data)

        if results:
            opts = {
                'leaf_offset': self.leaf_offset,
                'convert_c': self.convert_c,
                'now': self.now}
            measurements = Parser(**opts).parse(results)
            logger.debug("MEASUREMENTS: %s" % measurements)

            # send the measurements into the pipeline
            p = Pipeline()
            await p.input(measurements)

        return measurements


    async def collect_id(self, pk:int):
        await self.load_devices(pk)
        return await self.collect()


    def get_reader_list(self, devices:list) -> list:
        """
        parse list of devices and return only ble advertisement readers
        """

        results = []
        for d in devices:
            if (not hasattr(d.plugin, 'is_bluetooth') or
                d.plugin.is_bluetooth is not True):
                continue

            if (not hasattr(d.plugin, 'is_reader') or
                d.plugin.is_reader is not True):
                continue

            results.append(d)

        logger.debug("Reader List: %s" % results)
        return results


    def get_poller_list(self, devices:list) -> list:
        """
        parse list of devices and return only ble pollable clients
        """

        results = []
        for d in devices:
            if (not hasattr(d.plugin, 'is_bluetooth') or
                d.plugin.is_bluetooth is not True):
                continue

            if (not hasattr(d.plugin, 'is_pollable') or
                d.plugin.is_pollable is not True):
                continue

            results.append(d)

        logger.debug("Poller List: %s" % results)
        return results


    async def load_devices(self, ids:list = None):

        if ids is None:
            self.devices = await Device.select().where(Device.room_id != None)
        elif isinstance(ids, int):
            result = await Device.get_by_id(ids)
            self.devices = [result]
        elif isinstance(ids, list):
            self.devices = await Device.select().where(Device.id.in_(ids))

        return


    async def scan_advertised(self, devices:list, seconds:int = 40) -> dict:
        stop = asyncio.Event()
        timeout = self.now + datetime.timedelta(seconds=seconds)
        found = {d.id:None for d in devices}

        def callback(reporting, advertisement):
            addr = reporting.address
            for d in devices:
                try:
                    if found[d.id] is not None:
                        continue

                    if d.attributes['address'].upper() != addr.upper():
                        continue

                    results = d.read_advertisement(
                        reporting, advertisement)

                    if results is not None:
                        found[d.id] = results
                        break
                except:
                    pass

            if None not in found.values() and not stop.is_set():
                stop.set()


        while not stop.is_set() and datetime.datetime.now() < timeout:
            async with lock:
                async with bleak.BleakScanner(callback):
                    await asyncio.sleep(1)

        return found


    async def poll_clients(self, devices:list) -> dict:
        results = {d.id:None for d in devices}

        for d in devices:
            tries = 3
            success = False
            while tries and not success:
                logger.debug(f"Polling client {d.as_dict()}")
                data = None
                try:
                    async with lock:
                        data = await d.poll()
                        logger.debug(f"polled data: {data}")
                        if data is not None:
                            results[d.id] = data
                            success = True
                except Exception as x:
                    logger.warning(x)

                if not success:
                    tries -= 1

                await asyncio.sleep(random.random())

        return results



import datetime
from markupsafe import escape
from marshmallow import Schema, EXCLUDE, pre_load

class SafeSchema(Schema):
    """
    Subclass of the marshmallow Schema base.
    Excludes unknown fields, and sanitizes all as potential markdown data
    """

    class Meta:
        unknown = EXCLUDE

    @pre_load
    def _sanitize(self, data:dict, **kwargs):
        """
        sanitize all input data, with markupsafe.escape()
        """

        special_operators = ['>', '<', '>=', '<=', '==', '!=']
        for key, value in data.items():
            if not value:
                continue

            if isinstance(value, dict):
                value = self._sanitize(value)
                continue

            if (isinstance(value, int) or isinstance(value, float) or
                isinstance(value, datetime.datetime) or
                isinstance(value, list)):
                continue

            if (isinstance(value, str) and value in special_operators):
                continue

            clean = escape(value)
            data[key] = str(clean).strip()

        return data



import os
import unittest
import datetime
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.room import Room
from potnanny.models.action import Action, ActionSchema, ActionTrigger

CLEANUP = True
DBFILE = '/tmp/test.db'

async def init_tables():
    try:
        db.init(f'aiosqlite:///{DBFILE}', pragmas=(('foreign_keys', 1),))
        db.register(Room)
        db.register(Action)
        db.register(ActionTrigger)
        await Room.create_table()
        await Action.create_table()
        await ActionTrigger.create_table()
        r = await Room.create(name='test room')
        await r.save()
    except:
        pass


class TestModels(IsolatedAsyncioTestCase):
    @classmethod
    def tearDownClass(cls):
        if CLEANUP:
            try:
                os.unlink(DBFILE)
            except:
                pass

    async def asyncSetUp(self):
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        async with db.connection():
            obj = await Action.create(
                name='test action',
                room_id=1,
                interface='foo.bar.Baz',
                attributes={
                    'threshold': 'value gt 90'
                }
            )
            await obj.save()

        assert obj.id > 0
        assert isinstance(obj.as_dict(), dict)


    async def test_open_trigger(self):
        tag = 'test open trigger'

        a1 = await Action.create(
            name=f'action 1 ({tag})',
            room_id=1,
            interface='foo.bar.Baz1',
            attributes={
                'threshold': 'value gt 90'
            }
        )
        await a1.save()

        a2 = await Action.create(
            name=f'action 2 ({tag})',
            room_id=1,
            interface='foo.bar.Baz2',
            attributes={
                'threshold': 'value gt 90',
                'sleep_minutes': 1
            }
        )
        await a2.save()

        # create trigger assigned to action 2
        await a2.init_trigger()

        # check there IS NOT an open trigger
        truth1 = await a1.has_open_trigger()
        assert truth1 is False

        # check there IS an open trigger
        truth2 = await a2.has_open_trigger()
        assert truth2 is True




if __name__ == '__main__':
    unittest.main()



import os
import logging
from quart import Quart, redirect, url_for
from quart_wtf.csrf import CSRFProtect
from quart_auth import QuartAuth, Unauthorized
from potnanny.models.user import SessionUser


# globals #
BASEDIR = os.path.abspath(os.path.dirname(__file__))
auth_manager = QuartAuth()
auth_manager.user_class = SessionUser
logger = logging.getLogger(__name__)


def init_application(config):
    app = Quart(__name__, root_path=BASEDIR)
    configure_app(app)
    csrf = CSRFProtect(app)
    auth_manager.init_app(app)
    load_blueprints(app)
    load_views()
    load_handlers(app)
    return app


def configure_app(app):
    app.config['SECRET_KEY'] = os.getenv('FLASK_SECRET')
    app.config['WTF_CSRF_SECRET_KEY'] = os.getenv('FLASK_SECRET')
    app.config['WTF_CSRF_ENABLED'] = True
    app.config['EXPLAIN_TEMPLATE_LOADING'] = False

    ## development ##
    app.config['QUART_AUTH_COOKIE_SECURE'] = False


def load_blueprints(app):
    """
    Load flask blueprints from web apps

    args:
        - Flask app instance
    """

    from potnanny.api.environments.views import bp as environments
    from potnanny.api.rooms.views import bp as rooms
    from potnanny.api.devices.views import bp as devices
    from potnanny.api.keychains.views import bp as keychains
    from potnanny.api.graphs.views import bp as graphs
    from potnanny.api.utils.views import bp as utils
    from potnanny.api.settings.views import bp as settings
    from potnanny.api.controls.views import bp as controls
    from potnanny.api.features.views import bp as features
    from potnanny.api.auth.views import bp as auth
    from potnanny.api.about.views import bp as about
    from potnanny.api.errors.views import bp as errors
    from potnanny.api.actions.views import bp as actions
    from potnanny.api.plugins.views import bp as plugins

    app.register_blueprint(environments)
    app.register_blueprint(rooms)
    app.register_blueprint(devices)
    app.register_blueprint(keychains)
    app.register_blueprint(graphs)
    app.register_blueprint(utils)
    app.register_blueprint(settings)
    app.register_blueprint(controls)
    app.register_blueprint(features)
    app.register_blueprint(auth)
    app.register_blueprint(about)
    app.register_blueprint(errors)
    app.register_blueprint(actions)
    app.register_blueprint(plugins)


def load_views():
    """
    Load views from any of the web apps
    """

    from potnanny.api.environments import views
    from potnanny.api.rooms import views
    from potnanny.api.devices import views
    from potnanny.api.keychains import views
    from potnanny.api.graphs import views
    from potnanny.api.utils import views
    from potnanny.api.settings import views
    from potnanny.api.controls import views
    from potnanny.api.features import views
    from potnanny.api.auth import views
    from potnanny.api.about import views
    from potnanny.api.errors import views
    from potnanny.api.actions import views
    from potnanny.api.plugins import views



def load_handlers(app):
    @app.errorhandler(Unauthorized)
    async def redirect_to_login(Exception):
        return redirect(url_for("auth.login"))



import os
import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.room import Room

DBFILE = '/tmp/test.db'

async def init_tables():
    db.init(f'aiosqlite:///{DBFILE}', pragmas=(('foreign_keys', 1),))
    db.register(Room)
    await Room.create_table()
    async with db.transaction():
        r = await Room.get_or_create(name='test')

class TestDBFuncs(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        await init_tables()

    @classmethod
    def tearDownClass(cls):
        try:
            os.unlink(DBFILE)
        except:
            pass

    async def test_transaction_rollback(self):
        async with db.transaction() as trx:
            for i in [1,2,3,4]:
                r = await Room.create(name=f"room {i}")

            await trx.rollback()

        check = await Room.select().where(Room.name == "room 1").first()
        assert check is None


    async def test_transaction(self):
        async with db.transaction() as trx:
            for i in [10,11,12,13]:
                r = await Room.create(name=f"room {i}")

        check = await Room.select().where(Room.name == "room 10").first()
        assert check.name == 'room 10'


if __name__ == '__main__':
    unittest.main()



import logging
import datetime
import marshmallow
from peewee_aio import fields
from quart_auth import AuthUser
from potnanny.database import BaseModel
from potnanny.models.schemas.safe import SafeSchema


logger = logging.getLogger(__name__)


class UserSchema(SafeSchema):
    name = marshmallow.fields.String()
    password = marshmallow.fields.String(allow_none=True)
    roles = marshmallow.fields.String(allow_none=True)


class User(BaseModel):
    id = fields.AutoField()
    name = fields.CharField(24, unique=True)
    roles = fields.CharField(48, default='user')
    password = fields.CharField(512)
    created = fields.DateTimeField(default=datetime.datetime.utcnow)

    def __str__(self):
        return f"<User id={self.id}, name='{self.name}'>"

    def as_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'roles': self.notes,
            'created': self.created.isoformat() + "Z",
        }


class SessionUser(AuthUser):
    def __init__(self, auth_id):
        super().__init__(auth_id)
        self._resolved = False
        self._user = None

    async def _resolve(self):
        if not self._resolved:
            self._user = await User.get_by_id(int(self._auth_id))
            self._resolved = True

    @property
    async def name(self):
        await self._resolve()
        return self._user.name



import logging
import asyncio
import datetime
import marshmallow
from peewee_aio import fields
from potnanny.database import BaseModel
from potnanny.models.schemas.safe import SafeSchema
from potnanny.models.mixins import InterfaceMixin
from potnanny.utils.eval import evaluate
from .ext import JSONField


logger = logging.getLogger(__name__)


class ActionSchema(SafeSchema):
    name = marshmallow.fields.String()
    attributes = marshmallow.fields.Dict(allow_none=True)
    interface = marshmallow.fields.String()


class Action(BaseModel, InterfaceMixin):
    id = fields.AutoField()
    name = fields.CharField(48)
    interface = fields.CharField(64)
    attributes = JSONField(default={})
    created = fields.DateTimeField(default=datetime.datetime.utcnow)


    def __str__(self):
        return f"<Action id={self.id}, name='{self.name}'>"


    def as_dict(self) -> dict:
        return {
            'id': self.id,
            'name': self.name,
            'interface': self.interface,
            'attributes': self.attributes,
            'created': self.created.isoformat() + "Z",
        }


    async def input(self, measurement):
        # if existing acton trigger, we skip and wait for trigger to expire
        existing = await self.has_open_trigger()
        if existing:
            return

        stmt = "{0:0.1f} {1:s} {2:0.1f}".format(
            measurement['value'],
            self.attributes['condition'],
            self.attributes['threshold'])
        truth = evaluate(stmt)
        if not truth:
            return

        try:
            # do things here
            klass = self.interface_class(self.interface)
            inst = klass()
            await inst.input(measurement)

            # finally, we assume we did things with our action interface,
            # and now we should remember that something occurred
            await self.init_trigger()
        except Exception as x:
            logger.warning(x)


    async def has_open_trigger(self):
        now = datetime.datetime.utcnow()
        triggers = await ActionTrigger.select().where(
            ActionTrigger.expires > now,
            ActionTrigger.action_id == self.id
        )

        try:
            return bool(len(list(triggers)))
        except:
            return False


    async def init_trigger(self):
        try:
            now = datetime.datetime.utcnow()
            if 'sleep_minutes' in self.attributes:
                end = now + datetime.timedelta(
                    minutes=int(self.attributes['sleep_minutes']))
            else:
                end = now

            t = await ActionTrigger.create(
                created=now,
                expires=end,
                action_id=self.id
            )
            await t.save()
        except Exception as x:
            logger.warning(x)


class ActionTrigger(BaseModel):
    id = fields.AutoField()
    created = fields.DateTimeField(default=datetime.datetime.utcnow)
    expires = fields.DateTimeField(default=datetime.datetime.utcnow)
    action = fields.ForeignKeyField(Action,
        on_delete='CASCADE',
        backref='triggers')


    def __str__(self):
        return f"<ActionTrigger id={self.id}, expires={self.expires}>"


    def as_dict(self) -> dict:
        return {
            'id': self.id,
            'action_id': self.action_id,
            'created': self.created.isoformat() + "Z",
            'expires': self.expires.isoformat() + "Z",
        }



import logging
import datetime
import marshmallow
from peewee_aio import fields
from potnanny.database import BaseModel
from potnanny.models.schemas.safe import SafeSchema
from .ext import JSONField


logger = logging.getLogger(__name__)


class KeychainSchema(SafeSchema):
    name = marshmallow.fields.String()
    attributes = marshmallow.fields.Dict(allow_none=True)
    protected = marshmallow.fields.Boolean(allow_none=True)


class Keychain(BaseModel):
    id = fields.AutoField()
    name = fields.CharField(48, unique=True)
    attributes = JSONField(default={})
    protected = fields.BooleanField(default=False)
    created = fields.DateTimeField(default=datetime.datetime.utcnow)

    def __str__(self):
        return f"<Keychain id={self.id}, name='{self.name}'>"

    def as_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'attributes': self.attributes,
            'protected': self.protected,
            'created': self.created.isoformat() + "Z",
        }



import asyncio
import logging
import json
import potnanny.controllers.worker as worker
from quart import (Blueprint, jsonify, render_template, abort, request, flash,
    redirect)
from quart_auth import login_required
from jinja2 import TemplateNotFound
from potnanny.models.device import Device, DeviceSchema
from potnanny.models.room import Room
from potnanny.models.keychain import Keychain
from potnanny.database import db, lock
from potnanny.ble import lock as ble_lock
from potnanny.controllers.outlet import switch_device_outlet


logger = logging.getLogger(__name__)
bp = Blueprint('devices', __name__, template_folder='templates')


@bp.route('/devices', methods=['GET'])
@login_required
async def index():
    devices = await Device.select()
    try:
        features = await Keychain.select().where(Keychain.name == 'features')[0]
        logger.debug(features)
    except:
        features = None

    try:
        return await render_template('devices/index.html',
            data=devices, features=features)
    except TemplateNotFound:
        abort(404)


@bp.route('/api/v1.0/devices', methods=['GET'])
@login_required
async def get_list():
    objects = await Device.select()
    if not objects:
        return jsonify({"status": "error", "msg": "object not found"}), 404

    return jsonify({
        "status": "ok", "msg": [o.as_dict()  for o in objects] }), 200


@bp.route('/api/v1.0/devices', methods=['POST'])
@login_required
async def create():
    try:
        jsondata = await request.get_json()
        schema = DeviceSchema()
        data = schema.load(jsondata)
        async with lock:
            async with db.connection():
                obj = await Device.create(**data)
                await obj.save()
        return jsonify({"status": "ok", "msg": obj.as_dict()}), 201
    except Exception as x:
        return jsonify({"status": "error", "msg": str(x) }), 500


@bp.route('/api/v1.0/devices/<int:pk>', methods=['GET'])
@login_required
async def get(pk):
    obj = await Device.get_by_id(pk)
    if not obj:
        return jsonify(
            {"status": "error", "msg": f"object with id {pk} not found"}), 404

    return jsonify({"status": "ok", "msg": obj.as_dict()}), 200


@bp.route('/api/v1.0/devices/<int:pk>', methods=['PATCH'])
@login_required
async def patch(pk):
    try:
        jsondata = await request.get_json()
        schema = DeviceSchema()
        data = schema.load(jsondata)
        async with lock:
            async with db.connection():
                obj = await Device.get_by_id(pk)
                for k, v in data.items():
                    setattr(obj, k, v)
                await obj.save()

        return jsonify({"status": "ok", "msg": obj.as_dict()}), 200
    except Exception as x:
        logger.warning(str(x))
        return jsonify(
            {"status": "error", "msg": f"object with id {pk} not found"}), 404


@bp.route('/api/v1.0/devices/<int:pk>', methods=['DELETE'])
@login_required
async def delete(pk):
    async with lock:
        async with db.connection():
            try:
                obj = await Device.get_by_id(pk)
                await obj.delete_instance()
                await flash("Device deleted", "info")
            except:
                pass

    return jsonify({
        "status": "ok", "msg": ""}), 200


@bp.route('/api/v1.0/devices/<int:pk>/poll', methods=['POST'])
@login_required
async def device_poll(pk):
    asyncio.create_task(worker.WORKER.poll_device(pk))
    return jsonify({"status": "ok", "msg": "polling queued"}), 200


@bp.route('/api/v1.0/devices/<int:pk>/<int:outlet>/on', methods=['POST'])
@login_required
async def device_on(pk, outlet):
    state = 1
    result = await switch_device_outlet(pk, outlet, state)
    if result != state:
        return jsonify({
            "status": "error",
            "msg": f"error switching device {pk} outlet {outlet}"}), 500

    return jsonify({"status": "ok", "msg": {'state': state}}), 200


@bp.route('/api/v1.0/devices/<int:pk>/<int:outlet>/off', methods=['POST'])
@login_required
async def device_off(pk, outlet):
    state = 0
    result = await switch_device_outlet(pk, outlet, state)
    if result != state:
        return jsonify({
            "status": "error",
            "msg": f"error switching device {pk} outlet {outlet}"}), 500

    return jsonify({"status": "ok", "msg": {'state': state}}), 200


@bp.route('/api/v1.0/devices/<int:pk>/keycode', methods=['GET'])
@login_required
async def get_keycode(pk):
    device = await Device.get_by_id(pk)
    if not device:
        return jsonify({
            "status": "error", "msg": "not found"}), 404

    if (not hasattr(device.plugin, 'is_key_required') or
        device.plugin.is_key_required is not True):
        return jsonify({
            "status": "error",
            "msg": "not a valid request for this device"}), 405

    secret = None
    try:
        async with ble_lock:
            secret = await device.plugin.scan_key()
    except Exception as x:
        logger.debug(x)

    if secret:
        return jsonify({"status": "ok", "msg": secret}), 200
    else:
        return jsonify({
            "status": "error",
            "msg": "key code not detected before timeout"}), 404



from .base import (DevicePlugin, BluetoothDevicePlugin, GPIODevicePlugin,
    ActionPlugin, PipelinePlugin)

plugin_category_map = {
    'device': [DevicePlugin, BluetoothDevicePlugin, GPIODevicePlugin],
    'action': [ActionPlugin],
    'pipeline': [PipelinePlugin],
}



import asyncio
import logging
import datetime
from potnanny.database import db
from potnanny.models.measurement import Measurement


logger = logging.getLogger(__name__)


async def purge_measurements(cutoff:datetime.datetime):
    """
    Remove measurements from table when older than cutoff datetime
    """

    try:
        async with db.connection():
            result = await Measurement.delete().where(
                Measurement.created < cutoff)
    except Exception as x:
        logger.warning(x)



import logging
from quart import Blueprint, jsonify, render_template, abort
from quart_auth import login_required
from jinja2 import TemplateNotFound
from potnanny.models.keychain import Keychain, KeychainSchema
from potnanny.database import db


logger = logging.getLogger(__name__)
bp = Blueprint('settings', __name__, template_folder='templates')


@bp.route('/settings', methods=['GET'])
@login_required
async def index():
    obj = await Keychain.select().where(Keychain.name == 'settings')[0]

    try:
        return await render_template('settings/index.html', title='Settings',
            data=obj.as_dict())
    except TemplateNotFound:
        abort(404)


@bp.route('/api/v1.0/settings', methods=['GET'])
@login_required
async def get():
    try:
        obj = await Keychain.select().where(Keychain.name == 'settings')[0]
        if not obj:
            return jsonify({
                "status": "error", "msg": f"object id {pk} not found"}), 404

        return jsonify({"status": "ok", "msg": obj.as_dict()}), 200
    except Exception as x:
        return jsonify({"status": "error", "msg": str(x)}), 500


@bp.route('/api/v1.0/settings', methods=['PATCH'])
@login_required
async def patch():
    try:
        jsondata = await request.get_json()
        schema = KeychainSchema()
        data = schema.load(jsondata)
        async with db.connection():
            obj = await Keychain.select().where(Keychain.name == 'settings')[0]
            if not obj:
                return jsonify({
                    "status": "error", "msg": f"object id {pk} not found"}), 404

            for k, v in data.items():
                setattr(obj, k, v)
            await obj.save()

        return jsonify({"status": "ok", "msg": obj.as_dict()}), 200
    except Exception as x:
        return jsonify({"status": "error", "msg": str(x)}), 500



import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.device import Device
from potnanny.plugins.utils import load_plugins


async def init_tables():
    db.init('aiosqlite:////tmp/test.db')
    async with db.connection():
        await Device.create_table()


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        print("loading plugins...")
        load_plugins('~/potnanny/plugins')
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_setup(self):
        opts = {
            'name': 'SwitchBot Hygrometer',
            'interface': 'device.ble.switchbot_hygrometer.SwitchbotHygrometer',
            'attribute': {"address": "EB:B3:0E:C7:7C:D2"}
        }

        d = await Device.create(**opts)
        await d.save()
        print("device saved")

        print("device plugin details...")
        print(d.plugin)


if __name__ == '__main__':
    unittest.main()



import re
import logging
from quart import Blueprint, jsonify, render_template, abort, request
from quart_auth import login_required
from jinja2 import TemplateNotFound
from potnanny.models.keychain import Keychain, KeychainSchema
from potnanny.database import db
from potnanny.utils.crypt import encrypt_str


logger = logging.getLogger(__name__)
bp = Blueprint('keychains', __name__, template_folder='templates')


@bp.route('/keychains', methods=['GET'])
@login_required
async def get_view():
    objects = await Keychain.select().where(Keychain.protected == False)
    try:
        return await render_template('keychains/index.html',
            data=[o for o in objects])
    except TemplateNotFound:
        abort(404)


@bp.route('/api/v1.0/keychains', methods=['GET'])
@login_required
async def get_list():
    objects = await Keychain.select()
    if not objects:
        return jsonify({"status": "error", "msg": "object not found"}), 404

    return jsonify({
        "status": "ok", "msg": [o.as_dict()  for o in objects] }), 200


@bp.route('/api/v1.0/keychains', methods=['POST'])
@login_required
async def create():
    jsondata = await request.get_json()
    schema = KeychainSchema()
    data = schema.load(jsondata)
    try:
        if 'password' in data['attributes']:
            # does this look already encrypted?
            if not re.search(r'^g.+=$', data['attributes']['password']):
                pw = encrypt_str(data['attributes']['password'])
                data['attributes']['password'] = pw
    except:
        pass

    try:
        async with db.connection():
            obj = await Keychain.create(**data)
            await obj.save()

        await flash("Keychain created", "info")
        return jsonify({"status": "ok", "msg": obj.as_dict()}), 201
    except Exception as x:
        return jsonify({"status": "error", "msg": str(x) }), 500


@bp.route('/api/v1.0/keychains/<int:pk>', methods=['GET'])
@login_required
async def get(pk):
    obj = await Keychain.get_by_id(pk)
    if not obj:
        return jsonify(
            {"status": "error", "msg": f"object with id {pk} not found"}), 404

    return jsonify({"status": "ok", "msg": obj.as_dict()}), 200


@bp.route('/api/v1.0/keychains/<int:pk>', methods=['PATCH'])
@login_required
async def patch(pk):
    jsondata = await request.get_json()
    schema = KeychainSchema()
    data = schema.load(jsondata)
    obj = None

    try:
        if 'password' in data['attributes']:
            # does this look already encrypted?
            if not re.search(r'^g.+=$', data['attributes']['password']):
                pw = encrypt_str(data['attributes']['password'])
                data['attributes']['password'] = pw
    except:
        pass

    async with db.connection():
        obj = await Keychain.get_by_id(pk)
        if not obj:
            return jsonify(
                {"status": "error", "msg": f"object with id {pk} not found"}), 404

        for k,v in data.items():
            setattr(obj, k, v)
        await obj.save()

    return jsonify({"status": "ok", "msg": obj.as_dict()}), 200


@bp.route('/api/v1.0/keychains/<int:pk>', methods=['DELETE'])
@login_required
async def delete(pk):
    async with db.connection():
        try:
            obj = await Keychain.get_by_id(pk)
            if obj.protected == True:
                return jsonify({
                    "status": "error", "msg": "forbidden"}), 403

            await obj.delete_instance()
            await flash("Keychain deleted", "success")
        except:
            pass

    return jsonify({
        "status": "ok", "msg": ""}), 200



import logging
from quart import (Blueprint, render_template, redirect, jsonify, request,
    abort, flash)
from quart_auth import login_required
from jinja2 import TemplateNotFound
from potnanny.utils import iso_from_sqlite
from potnanny.models.room import Room, RoomSchema
from potnanny.database import db
from .forms import RoomForm


logger = logging.getLogger(__name__)
bp = Blueprint('rooms', __name__, template_folder='templates')


@bp.route('/rooms/edit', methods=['GET', 'POST'])
@bp.route('/rooms/edit/<int:pk>', methods=['GET', 'POST'])
@login_required
async def room_edit(pk:int = 0):
    if pk:
        obj = await Room.get_by_id(pk)
        if not obj:
            abort(404)
    else:
        obj = {}

    form = await RoomForm.create_form(obj=obj)

    if request.method == 'POST':
        logger.debug(form.__dict__)
        logger.debug(f"form delete value: '{form.remove.data}'")

        if await form.validate_on_submit():
            if pk:
                if form.remove.data == 'true':
                    async with db.connection():
                        await obj.delete_instance()
                    await flash("room deleted", 'info')
                else:
                    obj.name = form.name.data
                    obj.notes = form.notes.data
                    await obj.save()
                    await flash("room updated", 'info')
            else:
                room = await Room.create(
                    name=form.name.data,
                    notes=form.notes.data)
                await room.save()
                await flash(f"room '{room.name}' created", 'info')
            return redirect("/")

    try:
        return await render_template('rooms/form.html', form=form)
    except TemplateNotFound:
        abort(404)



@bp.route('/api/v1.0/rooms', methods=['GET'])
@login_required
async def get_list():
    objects = await Room.select()
    if not objects:
        return jsonify({"status": "error", "msg": "no data"}), 404

    payload = [o.as_dict() for o in objects]
    return jsonify({"status": "ok", "msg": payload}), 200


@bp.route('/api/v1.0/rooms', methods=['POST'])
@login_required
async def create():
    try:
        jsondata = await request.get_json()
        schema = RoomSchema()
        data = schema.load(jsondata)
        async with db.connection():
            obj = await Room.create(**data)
            await obj.save()

        await flash("Room created", "info")
        return jsonify({"status": "ok", "msg": obj.as_dict()}), 201
    except Exception as x:
        return jsonify({"status": "error", "msg": str(x)}), 500


@bp.route('/api/v1.0/rooms/<int:pk>', methods=['GET'])
@login_required
async def get(pk):
    payload = None
    obj = await Room.get_by_id(pk)
    if obj is None:
        return jsonify({
            "status": "error", "msg": f"object id {pk} not found"}), 404

    return jsonify({"status": "ok", "msg": obj.as_dict()}), 200


@bp.route('/api/v1.0/rooms/<int:pk>', methods=['PATCH'])
@login_required
async def patch(pk):
    try:
        logger.debug(request.__dict__)
        jsondata = await request.get_json()
        logger.debug(jsondata)
        schema = RoomSchema()
        data = schema.load(jsondata)
        async with db.connection():
            obj = await Room.get_by_id(pk)
            if not obj:
                return jsonify({
                    "status": "error", "msg": f"object id {pk} not found"}), 404
            for k, v in data.items():
                setattr(obj, k, v)
            await obj.save()

        return jsonify({"status": "ok", "msg": obj.as_dict()}), 200
    except Exception as x:
        logger.warning(x)
        return jsonify({"status": "error", "msg": str(x)}), 500


@bp.route('/api/v1.0/rooms/<int:pk>', methods=['DELETE'])
@login_required
async def delete(pk):
    async with db.connection():
        try:
            obj = await Room.get_by_id(pk)
            await obj.delete_instance()
            await flash("Room deleted", "info")
        except:
            pass

    return jsonify({"status": "ok", "msg": ""}), 202



import unittest
import datetime
import random
from unittest import IsolatedAsyncioTestCase
from peewee import fn
from potnanny.database import db, init_db
from potnanny.models.room import Room
from potnanny.models.device import Device
from potnanny.models.measurement import Measurement
from potnanny.utils import utcnow


async def init_tables():
    db.init('aiosqlite:////tmp/test.db', pragmas=(('foreign_keys', 1),))

    db.register(Room)
    db.register(Device)
    db.register(Measurement)

    async with db.connection():
        await Room.create_table()
        await Device.create_table()
        await Measurement.create_table()


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_sql_query(self):
        async with db.connection():
            for i in range(1,11):
                r = await Room.create(name=f"test room {i}")
                await r.save()

        async with db.connection():
            sql = "select * from room"
            results = await db.fetchall(sql)
            assert len(results) >= 10


    async def test_func_query(self):
        now = utcnow()
        async with db.connection():
            r = await Room.create(name='initial room')
            await r.save()

            d = await Device.create(name='initial device', room_id=r.id,
                interface='foo.bar.baz')
            await d.save()

            for i in range(0, 10):
                then = now - datetime.timedelta(minutes=i)
                opts = {
                    'type': 'temperature',
                    'value': random.randrange(65, 80) * 1.0,
                    'device_id': 1,
                    'created': then
                }
                m = await Measurement.create(**opts)
                await m.save()


        async with db.connection():
            query = await Measurement.select(
                Measurement.id,
                Measurement.value,
                Measurement.type,
                Measurement.device_id,
                fn.MAX(Measurement.created)
            ).where(
                Measurement.device_id == 1,
                Measurement.type == "temperature"
            ).group_by(
                Measurement.device_id
            )
            
            m = query[0]
            assert m.created == now



if __name__ == '__main__':
    unittest.main()



import os
import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.utils.crypt import get_fernet_key, encrypt_str, decrypt_str


PRIVATE_KEY = 'foobarbaz!1234567890'
os.environ['POTNANNY_SECRET'] = PRIVATE_KEY


class TestCryptUtils(IsolatedAsyncioTestCase):
    async def test_get_key(self):
        key = get_fernet_key()
        assert isinstance(key, bytes)


    async def test_encrypt(self):
        plain = 'foo bar baz'
        secret = encrypt_str(plain)
        assert isinstance(secret, str)
        assert secret.startswith('g')
        assert secret.endswith('=')


    async def test_decrypt(self):
        plain = 'foo bar baz'
        secret = encrypt_str(plain)
        decoded = decrypt_str(secret)
        assert decoded == plain


if __name__ == '__main__':
    unittest.main()



import os
import asyncio
import logging
import peewee
from peewee_aio import Manager, AIOModel, fields
from peewee_aio.databases import get_db
from peewee import DatabaseProxy, logger
from weakref import WeakSet
from potnanny.utils.password import hash_password, verify_password


# lock used for db inserts
lock = asyncio.Lock()
logger = logging.getLogger(__name__)


class Database(Manager):
    def __init__(self):
        self.url = None
        self.backend_options = {}
        self.pw_database = DatabaseProxy()

    def init(self, url:str, **backend_options):
        if not url and not self.url:
            raise ValueError("DB url was never given!")
        backend_options.setdefault("convert_params", True)
        self.backend_options.update(backend_options)
        super(Manager, self).__init__(url or self.url, logger=logger,
			**self.backend_options)

        self.models = WeakSet()
        self.pw_database.initialize(get_db(self))


db = Database()


class BaseModel(AIOModel):
    __rel__: {}
    _manager = db

    class Meta:
        database = db.pw_database


async def init_db(url:str, **kwargs):
	db.init(url, **kwargs)
	await init_tables()


async def init_tables():
    from .models.room import Room
    from .models.device import Device
    from .models.measurement import Measurement
    from .models.keychain import Keychain
    from .models.control import Control
    from .models.license import License
    from .models.user import User
    from .models.action import Action, ActionTrigger

    db.register(Room)
    db.register(Device)
    db.register(Measurement)
    db.register(Keychain)
    db.register(Control)
    db.register(License)
    db.register(User)
    db.register(Action)
    db.register(ActionTrigger)

    async with db.connection():
        await Room.create_table()
        await Device.create_table()
        await Measurement.create_table()
        await Keychain.create_table()
        await Control.create_table()
        await User.create_table()
        await License.create_table()
        await Action.create_table()
        await ActionTrigger.create_table()

        try:
            # create default settings
            s_opts = {
                'name': 'settings',
                'protected': True,
                'attributes': {
                    'temperature_display': 'F',
                    'polling_interval': 10,
                    'plugin_path': os.path.expanduser('~/potnanny/plugins'),
                    'leaf_offset': -2,
                    'storage_days': 7,
                    'graph_hours': 24,
                }
            }
            obj = await Keychain.create(**s_opts)
            await obj.save()
        except:
            pass

        try:
            # create default features
            f_opts = {
                'name': 'features',
                'protected': True,
                'attributes': {
                    'room_limit': 1,
                    'device_limit': 4,
                }
            }
            obj = await Keychain.create(**f_opts)
            await obj.save()
        except:
            pass

        try:
            # create default user
            u_opts = {
                'name': 'admin',
                'password': hash_password('potnanny!'),
                'protected': True,
                'roles': "user,admin"
            }
            u = await User.create(**u_opts)
            await u.save()
        except:
            pass



import unittest
import datetime
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.schedule import Schedule


DBFILE = '/tmp/test.db'

async def init_tables():
    db.init(f'aiosqlite:///{DBFILE}')
    db.register(Schedule)
    await Schedule.create_table()


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        await init_tables()

    @classmethod
    def tearDownClass(cls):
        try:
            os.unlink(DBFILE)
        except:
            pass

    async def test_create(self):
        opts = {
            'name': 'test',
            'days': 127,
            'on_time': '06:00',
            'off_time': '07:00',
            'device': 1,
            'outlet': 1
        }
        s = await Schedule.create(**opts)
        await s.save()
        assert s.id > 0


    async def test_runs(self):
        now = datetime.datetime.now().replace(second=0,microsecond=0)
        then = now + datetime.timedelta(minutes=10)

        opts = {
            'name': 'immediate test',
            'days': 127,
            'on_time': f"{now.hour}:{now.minute}",
            'off_time': "00:01",
            'device': 1,
            'outlet': 1
        }

        s = await Schedule.create(**opts)
        await s.save()

        # should always produce True
        true, key = s.runs_now(now, True)
        assert true is True
        assert key == 'on'

        # should produce false
        true, key = s.runs_now(then, True)
        assert true is False


if __name__ == '__main__':
    unittest.main()



import logging
from potnanny.plugins.category import plugin_category_map


logger = logging.getLogger(__name__)


async def get_plugin_map() -> dict:
    """
    Get dict of plugins mapped by category
    """

    data = {}
    for k in plugin_category_map.keys():
        if k not in data:
            data[k] = []

        for klass in plugin_category_map[k]:
            for p in klass.plugins:
                try:
                    data[k].append(extract_plugin_data(p))
                except Exception as x:
                    logger.warning(x)

    return data


def extract_plugin_data(p) -> dict:
    data = {'interface': '.'.join((p.__module__, p.__name__))}

    try:
        data['name']  = p.name
    except:
        pass
    try:
        data['description'] = p.description
    except:
        pass

    return data



"""
Functions for calculating Vapor Pressure Deficit (VPD)
Formulats derived from here:
    - http://eebweb.arizona.edu/faculty/saleska/SWES.410.510/LECTURES/Dominguez_Microphysics.pdf
"""

def calculate_vpd(t: float, h: float) -> float:
    """
    Calculate the Vapour Pressure Deficit from temp and %rh

    args:
        - Temperature (Celsius)
        - RH(%)
    returns:
        - Float (kPascals)
    """

    svp = calculate_svp(t)
    pa = svp * ((100 - h)/100)
    kpa = pa/1000
    return float("%0.2f" % kpa)


def calculate_svp(t: float) -> float:
    """
    Calculate Saturated Vapour Pressure from temperature

    args:
        - temperature (celsius)
    returns:
        - float (value in Pascals)
    """

    e = 2.7182818284590452353602875
    svp = 611 * e**(t/(t + 237.3) * 17.27)
    return svp




import logging
from quart import Blueprint, jsonify, render_template, abort
from quart_auth import login_required
from jinja2 import TemplateNotFound
from potnanny.models.keychain import Keychain
from .controllers import get_environments, get_room_environments


logger = logging.getLogger(__name__)
bp = Blueprint('environments', __name__, template_folder='templates')


SUFFIXES = {
    'temperature': '°',
    'humidity': '%',
    'battery': '%',
    'soil_moisture': '%',
}


@bp.route('/', methods=['GET'])
@login_required
async def dashboard():
    data = await get_environments()
    try:
        features = await Keychain.select().where(Keychain.name == 'features')[0]
        logger.debug(features)
    except:
        features = None

    try:
        return await render_template('environments/dash.html',
            data=data,
            suffixes=SUFFIXES,
            features=features.as_dict())
    except TemplateNotFound:
        abort(404)


@bp.route('/environments/<int:pk>', methods=['GET'])
@login_required
async def room_dashboard(pk):
    data = await get_room_environments(pk)
    try:
        return await render_template('environments/room.html',
            data=data, suffixes=SUFFIXES)
    except TemplateNotFound:
        abort(404)


@bp.route('/api/v1.0/environments', methods=['GET'])
@login_required
async def environment():
    data = await get_environments()
    if data:
        return jsonify({'status': 'ok', 'msg': data}), 200
    else:
        return jsonify({'status': 'error', 'msg': 'no data'}), 404


@bp.route('/api/v1.0/environments/<int:pk>', methods=['GET'])
@login_required
async def room_environment(pk):
    data = await get_room_environments(pk)
    if data:
        return jsonify({'status': 'ok', 'msg': data}), 200
    else:
        return jsonify({'status': 'error', 'msg': 'no data'}), 404



from quart_wtf import QuartForm
from wtforms.fields import TextAreaField
from wtforms import validators


class LicenseForm(QuartForm):
     license = TextAreaField(
        'License Key',
        validators=[
            validators.DataRequired('must provide a license key'),
        ]
    )



import json
from peewee import Field

class JSONField(Field):
    field_type = 'text'

    def db_value(self, value):
        try:
            result = json.dumps(value)
        except:
            result = None
        return result

    def python_value(self, value):
        try:
            result = json.loads(value)
        except:
            result = None
        return result



import re
from .shell import run

LOCAL_ADDRESS = None

async def load_local_address():
    global LOCAL_ADDRESS

    if not LOCAL_ADDRESS:
        in_block = False
        cmd = "ifconfig -a"
        (rval, stdout, stderr) = await run(cmd)
        if rval == 0:
            for line in stdout.splitlines():
                match = re.search(r'inet (\d+\.\d+\.\d+\.\d+)', line)
                if match and match.group(1) != '127.0.0.1':
                    LOCAL_ADDRESS = match.group(1)
                    break

    return LOCAL_ADDRESS



import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.device import Device


async def init_tables():
    db.init('aiosqlite:////tmp/test.db')
    async with db.connection():
        await Device.create_table()


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        async with db.connection():
            device = await Device.create(name='device 1',
                interface='foo.bar.baz')
            await device.save()
            assert device.id > 0

    async def test_update(self):
        async with db.connection():
            device = await Device.create(name='device 2',
                interface='foo.bar.baz')
            await device.save()

            device.attributes = {'foo': 'bar'}
            await device.save()
            assert device.attributes == {'foo': 'bar'}

    async def test_delete(self):
        async with db.connection():
            device = await Device.create(name='device 3',
                interface='foo.bar.baz')
            await device.save()
            pk = device.id
            await device.delete_instance()

            row = await Device.select().where(Device.name == 'device 3')
            assert len(row) == 0


if __name__ == '__main__':
    unittest.main()



import os
import sys
import time
import logging
import asyncio
import signal
import argparse
from daemon import DaemonContext
from potnanny.utils.pids import PIDFILE, is_running, pid_to_file
from potnanny.utils import resolve_path
from potnanny.config import Config
from potnanny.app import run_app


args = None
logger = None


def main():
    parse_args()
    if args.action == 'start':
        start()
    elif args.action == 'stop':
        stop()
    elif args.action == 'status':
        status()
    elif args.action == 'restart':
        restart()
    else:
        sys.stderr.write(f"Unrecognized action '{args.action}'\n")
        sys.exit(1)


def start():
    print("Sevice starting")
    pid = is_running()
    if pid:
        sys.stderr.write(f"Service already running (pid {pid})\n")
        sys.exit(1)
    start_service()


def stop(die_absent=True):
    print("Service stopping")
    pid = is_running()
    if not pid:
        sys.stderr.write("Service not running\n")
        if die_absent:
            sys.exit(1)
    else:
        try:
            # pgid = os.getpgid(pid)
            # os.killpg(pgid, signal.SIGTERM)
            os.kill(pid, signal.SIGKILL)
        except Exception as x:
            print(x)
        finally:
            os.unlink(PIDFILE)
        print("Service stopped")


def status():
    print("Checking service")
    pid = is_running()
    if pid:
        print(f"Service running (pid {pid})")
        sys.exit(0)
    else:
        sys.stderr.write("Service not running\n")
        sys.exit(1)


def restart():
    stop(False)
    time.sleep(2)
    start()


def start_service():
    opts = {
        'stdout': None,
        'stderr': None
    }
    if args.debug:
        opts.update({
            'stdout': sys.stdout,
            'stderr': sys.stderr
        })

    with DaemonContext(**opts):
        config = Config()
        pid_to_file(os.getpid())
        init_logging(config.log_path)
        asyncio.run(run_app(config))


def init_logging(path=None):
    """
    Initialize logging to file, or stdout
    """

    global logger
    logger = logging.getLogger('potnanny')

    fmt = '%(asctime)s|%(name)s|%(levelname)s|%(message)s'
    if path:
        path = resolve_path(path)

    if args.debug:
        level = logging.DEBUG
        logging.basicConfig(
            level=level,
            format=fmt,
            datefmt='%Y-%m-%d %H:%M:%S')
        logging.getLogger('aiosqlite').setLevel(logging.INFO)
        logging.getLogger('bleak').setLevel(logging.INFO)
    else:
        level = logging.WARNING
        logging.basicConfig(
            filename=path,
            level=level,
            format=fmt,
            datefmt='%Y-%m-%d %H:%M:%S')


def parse_args():
    global args
    parser = argparse.ArgumentParser(description='Potnanny worker and web service')
    parser.add_argument('action', choices=[
        'start', 'stop', 'restart', 'status'])
    parser.add_argument('-d','--debug', action='store_true',
        help='turn on debug messages')
    args = parser.parse_args()


if __name__ == '__main__':
    main()



from quart_wtf import QuartForm
from quart_auth import current_user
from wtforms.fields import (StringField, PasswordField, HiddenField)
from wtforms import validators
from wtforms.widgets import PasswordInput


class LoginForm(QuartForm):
    username = StringField(
        'Username',
        validators=[
            validators.DataRequired('please enter username'),
        ]
    )

    password = PasswordField(
        'Password',
        widget=PasswordInput(hide_value=False),
        validators=[
            validators.DataRequired('please enter password'),
        ]
    )


class PasswordResetForm(QuartForm):
    username = HiddenField(
        validators=[validators.optional()]
    )
    current = PasswordField(
        'Current Password',
        widget=PasswordInput(hide_value=False),
        validators=[
            validators.DataRequired('please enter current password'),
        ]
    )
    password = PasswordField(
        'New Password',
        widget=PasswordInput(hide_value=False),
        validators=[
            validators.DataRequired('please enter a new password'),
        ]
    )
    again = PasswordField(
        'New Password again',
        widget=PasswordInput(hide_value=False),
        validators=[
            validators.DataRequired('passwords must match'),
        ]
    )



import logging
import aiofiles
import datetime


logger = logging.getLogger(__name__)


async def get_error_list(path:str, hours:int = 24, reverse:bool = True) -> list:
    """
    Get errors for the last N hours, from a named logfile
    """

    errors = []
    now = datetime.datetime.now()
    try:
        async with aiofiles.open(path, mode='r') as f:
            async for line in f:
                atoms = line.strip().split('|')
                if len(atoms) < 4:
                    continue

                if atoms[2] in ['INFO', 'DEBUG']:
                    continue

                dt = datetime.datetime.strptime(atoms[0], '%Y-%m-%d %H:%M:%S')
                if dt < now - datetime.timedelta(hours=hours):
                    continue

                errors.append({
                    'created': atoms[0],
                    'service': atoms[1],
                    'level': atoms[2],
                    'message': atoms[3]
                })
    except Exception as x:
        pass

    if reverse:
        errors.reverse()

    return errors



import datetime


def utcnow(dt=None):
    """
    Get a UTC datetime obj, but with second and microsecond field set to zero.
    This should be used for all measurement datetime objects in potnanny.

    args:
        datetime obj (optional)
    returns:
        utc datetime obj
    """

    if dt is None:
        dt = datetime.datetime.utcnow()

    return dt.replace(second=0, microsecond=0)


def datetime_from_iso(iso):
    """
    Get datetime object from a javascript ISO Date string
    args:
        - str, like '2021-03-14T00:24:41.123Z'
    returns:
        datetime obj
    """

    try:
        # try utc tz conversion first
        return datetime.datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S.%f%z")
    except:
        pass

    try:
        # try utc tz conversion first
        return datetime.datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S.%fZ")
    except:
        pass

    try:
        # try timestamp without tz
        return datetime.datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S.%f")
    except:
        pass

    try:
        # try timestamp without tz
        return datetime.datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S")
    except:
        pass

    return None


def datetime_from_sqlite(txt):
    """
    Get datetime object from a sqlite formatted datetime str
    args:
        - str, like '2021-03-14 00:24:41.123'
    returns:
        datetime obj
    """

    try:
        return datetime.datetime.strptime(txt, "%Y-%m-%d %H:%M:%S.%f")
    except:
        pass

    try:
        return datetime.datetime.strptime(txt, "%Y-%m-%d %H:%M:%S")
    except:
        pass

    return None


def iso_from_sqlite(txt, utc=True):
    """
    Get iso repr of datetime object from a sqlite formatted datetime str
    args:
        - str, like '2021-03-14 00:24:41.123'
        - utc (is the sqlite datetime string in UTC)
    returns:
        iso datetime str, like '2021-03-28T00:24:41.123Z'
    """

    dt = datetime_from_sqlite(txt)
    if dt:
        if utc:
            return dt.isoformat() + "Z"
        else:
            return dt.isoformat()
    else:
        return None



import datetime
from potnanny.utils import iso_from_sqlite
from potnanny.database import db

async def get_environments() -> dict:
    """
    Get average Temp/Humidity/VPD readings for all rooms
    """

    stmt = """
        SELECT id,name,type,AVG(value),created FROM
            (SELECT id, name FROM room) AS a
            LEFT JOIN (
                SELECT
                    d.id as device_id,
                    d.room_id,
                    m.type,
                    m.value,
                    MAX(m.created) as created
                FROM
                    measurement AS m,
                    device AS d
                WHERE
                    m.device_id = d.id
                AND
                    m.type IN ('temperature', 'humidity', 'vpd')
                GROUP BY
                    d.room_id,
                    d.id,
                    m.type
            ) AS b
            ON a.id = b.room_id
            GROUP BY
                a.id,
                b.type
    """

    data = {}
    rows = []
    async with db.connection():
        rows = await db.fetchall(stmt)

    # reformat the row data
    for row in rows:
        if row[0] not in data:
            data[row[0]] = {
                'id': row[0],
                'name': row[1],
                'measurements': {} }

        if row[3] is not None:
            data[row[0]]['measurements'].update({ row[2]: row[3] })
            data[row[0]]['created'] = iso_from_sqlite(row[4])

    return [data[r] for r in sorted(data.keys())]


async def get_room_environments(pk:int) -> dict:
    """
    Get Device readings for all devices in a room
    """

    stmt = """
        SELECT * FROM
            (SELECT
                r.id AS room_id,
                r.name AS room_name,
                d.id as device_id,
                d.name as device_name
             FROM
                room as r,
                device as d
             WHERE
                r.id = %d
             AND
                d.room_id = %d) AS a
            LEFT JOIN (
                SELECT
                    m.device_id as m_device_id,
                    m.type,
                    m.value,
                    MAX(m.created)
                FROM
                    measurement AS m
                GROUP BY
                    m.device_id,
                    m.type
            ) AS b
            ON a.device_id = b.m_device_id
            ORDER BY
                a.device_id
    """ % (pk, pk)

    data = {}
    rows = []
    async with db.connection():
        rows = await db.fetchall(stmt)

    # reformat the row data
    for row in rows:
        if 'devices' not in data:
            data = {'id': row[0], 'name': row[1], 'devices': {}}

        if row[2] not in data['devices']:
            data['devices'][row[2]] = {
                'id': row[2],
                'name': row[3],
                'measurements': {} }

        if row[5] is not None:
            data['devices'][row[2]]['measurements'].update({ row[5]: row[6] })
            data['devices'][row[2]]['created'] = iso_from_sqlite(row[7])

    data['devices'] = [data['devices'][k] for k in sorted(data['devices'].keys())]
    return data



import asyncio
import logging
import json
from quart import (Blueprint, jsonify, render_template, abort, request, flash,
    redirect)
from quart_auth import login_required
from jinja2 import TemplateNotFound
from .controllers import get_error_list


logger = logging.getLogger(__name__)
bp = Blueprint('errors', __name__, template_folder='templates')
try:
    ERR_PATH = logging.getLoggerClass().root.handlers[0].baseFilename
except:
    ERR_PATH = None


@bp.route('/errors', methods=['GET'])
@login_required
async def index():
    hours = 24
    errors = await get_error_list(ERR_PATH, hours, True)
    try:
        return await render_template('errors/index.html',
            data=errors,
            title=f"Errors from last {hours} hours")
    except TemplateNotFound:
        abort(404)


@bp.route('/api/v1.0/errors', methods=['GET'])
@login_required
async def get_list():
    errors = await get_error_list(ERR_PATH, 24, True)
    return jsonify({
        "status": "ok", "msg": errors}), 200



import os
import base64
from cryptography.fernet import Fernet


def get_fernet_key() -> bytes:
    required_size = 32

    k = os.getenv('POTNANNY_SECRET')
    if k is None:
        raise ValueError('POTNANNY_SECRET key not found')

    if len(k) > required_size:
        k = k[:required_size]
    elif len(k) < 32:
        k = k.zfill(required_size)

    return base64.urlsafe_b64encode(k.encode())


def encrypt_str(data:str) -> str:
    key = get_fernet_key()
    f = Fernet(key)

    if isinstance(data, str):
        data = data.encode()

    return f.encrypt(data).decode()


def decrypt_str(data:str) -> str:
    key = get_fernet_key()
    f = Fernet(key)

    if isinstance(data, str):
        data = data.encode()

    return f.decrypt(data).decode()



import asyncio
import logging
from potnanny.database import db
from potnanny.models.keychain import Keychain, KeychainSchema
from potnanny.models.license import License, LicenseSchema


logger = logging.getLogger(__name__)


async def load_license(data:dict):
    try:
        schema = LicenseSchema()
        clean = schema.load(data)
    except:
        return (1, "schema validation failure")

    async with db.transaction() as trx:
        try:
            license = await License.create(**clean)
            if 'actions' in data:
                if 'keychain_updates' in data['actions']:
                    for item in data['actions']['keychain_updates']:
                        kc = await Keychain.select().where(
                            Keychain.name == item['name']).first()
                        if kc:
                            kc.attributes.update(item['attributes'])
                            await kc.save()

                if 'keychain_creates' in data['actions']:
                    for item in data['actions']['keychain_creates']:
                        kc = await Keychain.create(**item)
                        await kc.save()

                if 'downloads' in data['actions']:
                    for item in data['actions']['downloads']:
                        pass

        except Exception as x:
            logger.warning(x)
            await trx.rollback()
            return (2, 'license insert failure')

    return (0, 'ok')



import asyncio
import logging
import potnanny.database as database
from potnanny.config import Config
from potnanny.api import init_application
from potnanny.controllers.worker import (WORKER_STOP, run_worker,
    restart_worker, stop_worker)
from potnanny.utils import resolve_path, load_serial_number


# globals #
logger = logging.getLogger(__name__)
WEB_TASK = None
WRKR_TASK = None


async def run_app(config=Config()):
    """
    Initialize and run the app
    """
    global WEB_TASK
    global WRKR_TASK

    # load sn into global
    await load_serial_number()

    # init db and tables
    await database.init_db(config.database_uri, pragmas=(('foreign_keys','1'),))

    # the quart app
    application = init_application(config)

    # tasks to serve web with hypercorn, and run worker process
    WEB_TASK = asyncio.create_task(application.run_task(port=8080))
    WRKR_TASK = asyncio.create_task(restart_worker())

    logger.debug(f"web task: {WEB_TASK}")
    logger.debug(f"worker task: {WRKR_TASK}")
    logger.debug("entering event loop sleep")

    while True:
        await asyncio.sleep(1)



import logging
import base64
import json
from quart import (Blueprint, jsonify, render_template, abort, flash, request,
    redirect)
from quart_auth import login_required
from jinja2 import TemplateNotFound
from potnanny.database import db
from potnanny.models.keychain import Keychain, KeychainSchema
from potnanny.models.license import License
from .forms import LicenseForm
from .controllers import load_license


logger = logging.getLogger(__name__)
bp = Blueprint('features', __name__, template_folder='templates')


@bp.route('/features', methods=['GET'])
@login_required
async def show_features():
    results = await License.select()

    try:
        return await render_template('features/index.html', data=results)
    except TemplateNotFound:
        abort(404)


@bp.route('/features/import', methods=['GET', 'POST'])
@login_required
async def import_key():
    form = await LicenseForm.create_form()

    if await form.validate_on_submit():
        await request.get_data()
        raw = (await request.form)["license"]
        data = {}

        try:
            b = base64.b64decode(raw)
            logger.debug(b)

            data = json.loads(b)
            logger.debug(data)
            rval, msg = await load_license(data)
            if rval:
                if not getattr(form.license, 'errors'):
                    form.license.errors = []
                form.license.errors.append(msg)

        except Exception as x:
            logger.warning(x)
            if not getattr(form.license, 'errors'):
                form.license.errors = []
            form.license.errors.append('error decoding key')

        if not getattr(form.license, 'errors'):
            await flash('feature license loaded', 'info')
            return redirect("/")
    try:
        return await render_template('features/form.html', form=form)
    except TemplateNotFound:
        abort(404)



import asyncio
import logging
import json
from quart import (Blueprint, jsonify, render_template, abort, request, flash,
    redirect)
from quart_auth import login_required
from jinja2 import TemplateNotFound
from .controllers import get_plugin_map


logger = logging.getLogger(__name__)
bp = Blueprint('plugins', __name__, template_folder='templates')


@bp.route('/plugins', methods=['GET'])
@login_required
async def index():
    data = await get_plugin_map()
    try:
        return await render_template('plugins/index.html',
            data=data,
            title=f"Installed Plugins")
    except TemplateNotFound:
        abort(404)


@bp.route('/api/v1.0/plugins', methods=['GET'])
@login_required
async def plugin_list():
    try:
        data = await get_plugin_map()
        return jsonify({
            "status": "ok", "msg": data }), 200
    except:
        return jsonify({
            "status": "error", "msg": 'no data' }), 404



from quart_wtf import QuartForm
from wtforms.fields import (StringField, TextAreaField, HiddenField)
from wtforms import validators


class RoomForm(QuartForm):
    id = HiddenField('id', validators=[validators.optional()])
    remove = HiddenField('remove', validators=[validators.optional()])
    name = StringField('name', validators=[validators.DataRequired()])
    notes = TextAreaField('notes', validators=[validators.optional()])



def flatten_list(mylist):
    """
    Recursively flatten a list which may contain other lists
    args:
        - list
    returns:
        - a new list
    """

    results = []
    for i in mylist:
        if type(i) is list:
            results += flatten_list(i)
        else:
            results.append(i)

    return results



import logging
import datetime
import marshmallow
from peewee_aio import fields
from potnanny.database import BaseModel
from potnanny.models.schemas.safe import SafeSchema


logger = logging.getLogger(__name__)


class RoomSchema(SafeSchema):
    name = marshmallow.fields.String()
    notes = marshmallow.fields.String(allow_none=True)


class Room(BaseModel):
    id = fields.AutoField()
    name = fields.CharField(48)
    notes = fields.TextField(null=True)
    created = fields.DateTimeField(default=datetime.datetime.utcnow)

    def __str__(self):
        return f"<Room id={self.id}, name='{self.name}'>"

    def as_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'notes': self.notes,
            'created': self.created.isoformat() + "Z",
        }



import logging
from quart import Blueprint, jsonify, request
from quart_auth import login_required
from .controllers import room_graph, device_graph


logger = logging.getLogger(__name__)
bp = Blueprint('graphs', __name__)


@bp.route('/api/v1.0/rooms/<int:pk>/graph', methods=['GET'])
@login_required
async def room(pk):
    try:
        hours = int(request.args.get('hours'))
    except:
        hours = 12

    try:
        mtype = request.args.get('measurement')
    except:
        mtype = None

    graph = await room_graph(pk, mtype, hours)
    if not graph:
        return jsonify({
            "status": "error", "msg": "no data"}), 404

    return jsonify({"status": "ok", "msg": graph}), 200


@bp.route('/api/v1.0/devices/<int:pk>/graph', methods=['GET'])
@login_required
async def device(pk):
    try:
        hours = int(request.args.get('hours'))
    except:
        hours = 12

    try:
        mtype = request.args.get('measurement')
    except:
        mtype = None

    logger.debug(f"devices graph measurement params: {pk} {mtype} {hours}")

    graph = await device_graph(pk, mtype, hours)
    if not graph:
        return jsonify({
            "status": "error", "msg": "no data"}), 404

    return jsonify({"status": "ok", "msg": graph}), 200



import re


def evaluate(equation: str) -> bool:
    """
    evaluate a conditional equation str, like:
    ("22 >= 80")        # False
    ("11.1 == 13.9")    # False
    ("2 eq 2.0")        # True
    """

    atoms = re.split(r'\s+', equation)
    if len(atoms) != 3:
        raise ValueError("Poorly formed equation '%s'" % equation)


    a = float(atoms[0])
    b = float(atoms[2])
    oper = atoms[1]

    if oper in ['lt', '<', '&lt;'] and a < b:
        return True
    elif oper in ['le', '<=', '&lt;='] and a <= b:
        return True
    elif oper in ['gt', '>', '&gt;'] and a > b:
        return True
    elif oper in ['ge', '>=', '&gt;='] and a >= b:
        return True
    elif oper in ['eq', '==', '='] and a == b:
        return True
    elif oper in ['ne', '!='] and a != b:
        return True

    return False



import asyncio
import logging
import json
from quart import (Blueprint, jsonify, render_template, abort, request, flash,
    redirect)
from quart_auth import login_required
from jinja2 import TemplateNotFound
from potnanny.models.action import Action, ActionSchema
from potnanny.api.plugins.controllers import get_plugin_map
from potnanny.database import db, lock


logger = logging.getLogger(__name__)
bp = Blueprint('actions', __name__, template_folder='templates')


@bp.route('/actions', methods=['GET'])
@login_required
async def index():
    actions = await Action.select()

    try:
        return await render_template('actions/index.html',
            data=actions)
    except TemplateNotFound:
        abort(404)


@bp.route('/api/v1.0/actions', methods=['GET'])
@login_required
async def get_list():
    objects = await Action.select()
    if not objects:
        return jsonify({"status": "error", "msg": "object not found"}), 404

    return jsonify({
        "status": "ok", "msg": [o.as_dict()  for o in objects] }), 200


@bp.route('/api/v1.0/actions', methods=['POST'])
@login_required
async def create():
    try:
        jsondata = await request.get_json()
        schema = ActionSchema()
        data = schema.load(jsondata)
        async with lock:
            async with db.connection():
                obj = await Action.create(**data)
                await obj.save()
        return jsonify({"status": "ok", "msg": obj.as_dict()}), 201
    except Exception as x:
        return jsonify({"status": "error", "msg": str(x) }), 500


@bp.route('/api/v1.0/actions/<int:pk>', methods=['GET'])
@login_required
async def get(pk):
    obj = await Action.get_by_id(pk)
    if not obj:
        return jsonify(
            {"status": "error", "msg": f"object with id {pk} not found"}), 404

    return jsonify({"status": "ok", "msg": obj.as_dict()}), 200


@bp.route('/api/v1.0/actions/<int:pk>', methods=['PATCH'])
@login_required
async def patch(pk):
    try:
        jsondata = await request.get_json()
        schema = ActionSchema()
        data = schema.load(jsondata)
        async with lock:
            async with db.connection():
                obj = await Action.get_by_id(pk)
                for k, v in data.items():
                    setattr(obj, k, v)
                await obj.save()

        return jsonify({"status": "ok", "msg": obj.as_dict()}), 200
    except Exception as x:
        logger.warning(str(x))
        return jsonify(
            {"status": "error", "msg": f"object with id {pk} not found"}), 404


@bp.route('/api/v1.0/actions/<int:pk>', methods=['DELETE'])
@login_required
async def delete(pk):
    async with lock:
        async with db.connection():
            try:
                obj = await Action.get_by_id(pk)
                await obj.delete_instance()
                await flash("Action deleted", "info")
            except:
                pass

    return jsonify({
        "status": "ok", "msg": ""}), 200



import logging
from quart import (Blueprint, jsonify, render_template, abort, flash, request,
    redirect)
from quart_auth import AuthUser, login_user, logout_user, login_required
from jinja2 import TemplateNotFound
from potnanny.models.user import User, UserSchema, SessionUser
from potnanny.utils.password import verify_password, hash_password
from .forms import LoginForm, PasswordResetForm


logger = logging.getLogger(__name__)
bp = Blueprint('auth', __name__, template_folder='templates')


@bp.route('/login', methods=['GET', 'POST'])
async def login():
    form = await LoginForm.create_form()

    if await form.validate_on_submit():
        await request.get_data()
        username = (await request.form)["username"]
        password = (await request.form)["password"]

        try:
            u = await User.select().where(User.name == username).first()
            if u:
                if verify_password(password, u.password):
                    s = SessionUser(u.id)
                    s._user = u
                    s._resolved = True
                    login_user(s)
                    return redirect('/')
                else:
                    form.username.errors = ["login failure"]
            else:
                form.username.errors = ["login failure"]
        except Exception as x:
            logger.debug(str(x))
            form.username.errors = ["unexpected error"]

    try:
        return await render_template('auth/login.html', form=form)
    except TemplateNotFound:
        abort(404)


@bp.route('/logout', methods=['GET', 'POST'])
@login_required
async def logout():
    logout_user()
    return redirect("/")


@bp.route('/pwreset', methods=['GET', 'POST'])
@login_required
async def pwreset():
    form = await PasswordResetForm.create_form()

    if await form.validate_on_submit():
        await request.get_data()
        username = (await request.form)["username"]
        existing = (await request.form)["current"]
        password = (await request.form)["password"]

        try:
            u = await User.select().where(User.name == username).first()
            if u:
                if verify_password(existing, u.password):
                    u.password = hash_password(password)
                    await u.save()
                    await flash('password successfully changed', 'success')
                    return redirect('/')
                else:
                    form.current.errors = ["login failure"]
            else:
                form.current.errors = ["user error"]
        except Exception as x:
            form.current.errors = ["unexpected error"]

    try:
        return await render_template('auth/reset.html', form=form)
    except TemplateNotFound:
        abort(404)



import logging
import datetime
import marshmallow
from peewee_aio import fields
from potnanny.database import BaseModel
from potnanny.models.schemas.safe import SafeSchema
from potnanny.utils import utcnow
from .device import Device


logger = logging.getLogger(__name__)


class MeasurementSchema(SafeSchema):
    type = marshmallow.fields.String(allow_none=False)
    value = marshmallow.fields.Float(allow_none=False)
    created = marshmallow.fields.Raw(allow_none=True)
    device_id = marshmallow.fields.Integer(allow_none=False)


class Measurement(BaseModel):
    id = fields.AutoField()
    type = fields.CharField(24)
    value = fields.FloatField()
    created = fields.DateTimeField(default=utcnow)
    device = fields.ForeignKeyField(Device,
        on_delete='CASCADE',
        backref='measurements' )

    def __str__(self):
        return f"<Measurement id={self.id}, type={self.type}, value={self.value}>"

    def as_dict(self):
        return {
            'id': self.id,
            'type': self.type,
            'value': self.value,
            'device_id': self.device_id,
            'created': self.created.isoformat() + "Z",
        }



import logging
from quart import Blueprint, render_template, jsonify, request
from quart_auth import login_required
from jinja2 import TemplateNotFound
from potnanny.controllers.discover import discover_new_devices
from potnanny.controllers.worker import restart_worker
from potnanny.utils.serial import load_serial_number
from potnanny.utils.network import load_local_address


logger = logging.getLogger(__name__)
bp = Blueprint('utils', __name__)


@bp.route('/api/v1.0/scan', methods=['GET'])
@login_required
async def scan_devices():
    results = await discover_new_devices()
    if not results:
        return jsonify({"status": "error", "msg": "unexpected failure"}), 404

    return jsonify({"status": "ok", "msg": results}), 200


@bp.route('/api/v1.0/restartworker', methods=['POST'])
@login_required
async def restart():
    results = await restart_worker()
    return jsonify({"status": "ok", "msg": "restarted"}), 200


@bp.route('/api/v1.0/serial', methods=['GET'])
@login_required
async def serial():
    result = await load_serial_number()
    return jsonify({"status": "ok", "msg": result}), 200


@bp.route('/api/v1.0/address', methods=['GET'])
@login_required
async def get_address():
    result = await load_local_address()
    return jsonify({"status": "ok", "msg": result}), 200



import copy
import logging
from potnanny.plugins.base import (BluetoothDevicePlugin, GPIODevicePlugin,
    ActionPlugin, PipelinePlugin)
from potnanny.models.keychain import Keychain


logger = logging.getLogger(__name__)


class InterfaceMixin:
    """
    Many Models have an "interface" attribute, which is a string reference to
    a plugin class. This mixin searches for the actual plugin class
    """

    @classmethod
    def interface_class(cls, name:str):
        klass = None
        parents = [BluetoothDevicePlugin, GPIODevicePlugin,
            ActionPlugin, PipelinePlugin]

        for p in parents:
            klass = p.get_named_class(name)
            if klass is not None:
                break

        return klass


class PluginMixin(InterfaceMixin):
    """
    Class for interfacing with Plugin object instance via the $parent.plugin property
    """

    @property
    def plugin(self):
        """
        Attribute to fetch an instantiated copy of our interfaces plugin.

        Required object attributes:
            - interface = like 'input.ble.govee.Hygrometer'
            - attributes = like {'address': '11:22:33:44:55:66'}

        Now, can communicate with the instance directly with initialized
        instance with self.plugin
        """

        if not hasattr(self, '_plugin'):
            self._plugin = None

        try:
            klass = self.interface_class(self.interface)
            self._plugin = klass(**self.attributes)
        except Exception as x:
            logger.warning(str(x))
            pass

        try:
            # set up plugin specific introspection attributes, like:
            #   - is_bluetooth,
            #   - is_gpio,
            #   - is_reader,
            #   - is_pollable,
            #   - is_switchable,
            #   - is_key_required, etc

            if hasattr(self._plugin, 'poll') and callable(self._plugin.poll):
                setattr(self._plugin, 'is_pollable', True)

            if isinstance(self._plugin, BluetoothDevicePlugin):
                setattr(self._plugin, 'is_bluetooth', True)
                if (hasattr(self._plugin, 'read_advertisement') and
                    callable(self._plugin.read_advertisement)):
                    setattr(self._plugin, 'is_reader', True)
            elif isinstance(self._plugin, GPIODevicePlugin):
                setattr(self._plugin, 'is_gpio', True)

            if (hasattr(self._plugin, 'set_state') and
                callable(self._plugin.set_state)):
                setattr(self._plugin, 'is_switchable', True)

            if hasattr(self._plugin, 'key_code'):
                setattr(self._plugin, 'is_key_required', True)
        except:
            pass

        return self._plugin


class KeychainMixin:
    """
    A mixin class for a plugin (especially plugins that inherit from the
    ActionPlugin class) that must be associated with a named keychain.
    This requires the plugin has a self._keychain_name attribute.
    The keychain must be loaded async during setup:
        await self._load_my_keychain()

    Then, the plugin can access it's keychain data like:
        self.keychain.attributes
    """

    async def _load_my_keychain(self):
        if not hasattr(self, '_keychain'):
            self._keychain = None

        if not hasattr(self, '_keychain_name'):
            self._keychain_name = ''

        if self._keychain is None:
            results = await Keychain.select().where(
                Keychain.name == self._keychain_name)
            if results:
                self._keychain = results[0]

    @property
    def keychain(self):
        return self._keychain


class DeviceMixin(PluginMixin):
    """
    Add ability for device to interact with plugins/controllers
    """

    async def poll(self):
        """
        Poll our plugin instance for measurements, if possible
        """

        if (not hasattr(self.plugin, 'is_pollable') or
            self.plugin.is_pollable is not True):
            return None

        results = {
            'id': self.id,
            'name': self.name,
            'values': {},
        }

        try:
            values = await self.plugin.poll()
            results.update({'values': values})
        except Exception as x:
            logger.warning(str(x))

        return results


    def read_advertisement(self, device, advertisement):
        """
        Use plugin instance to decode the BLE advertisement data
        args:
            the device reporting data
            the advertisement data
        returns:
            dict, or None
        """

        if (not hasattr(self.plugin, 'is_reader') or
            self.plugin.is_reader is not True):
            return None

        results = {
            'id': self.id,
            'name': self.name,
            'values': {},
        }

        try:
            values = self.plugin.read_advertisement(device, advertisement)
            if values:
                results['values'] = values
        except:
            pass

        return results


    async def on(self, outlet:int = 1):
        """
        Switch device ON
        """

        if (not hasattr(self.plugin, 'is_switchable') or
            self.plugin.is_switchable is not True):
            return None

        result = await self.plugin.on(outlet)
        return result


    async def off(self, outlet:int = 1):
        """
        Switch device OFF
        """

        if (not hasattr(self.plugin, 'is_switchable') or
            self.plugin.is_switchable is not True):
            return None

        result = await self.plugin.off(outlet)
        return result


    async def set_value(self, data:dict):
        """
        Set key/value data on device (future?).
        """

        if (not hasattr(self.plugin, 'is_settable') or
            self.plugin.is_settable is not True):
            return None

        if type(data) is not dict:
            logger.warning(f"Expecting dict but got {type(data)}")
            return None

        try:
            result = await self.plugin.set_value(data)
            return result
        except Exception as x:
            logger.warning(str(x))
            return None



import os
import logging
import yaml
from types import SimpleNamespace
from potnanny.models.keychain import Keychain
from potnanny.utils import resolve_path
from potnanny.database import db


logger = logging.getLogger(__name__)
DB_PATH = resolve_path('~/potnanny/potnanny.db')
DEFAULTS = {
    'database_uri': 'aiosqlite:///' + DB_PATH,
    'log_path': resolve_path('~/potnanny/errors.log'),
    'plugin_path': resolve_path('~/potnanny/plugins'),
}


class Config:
    def __init__(self, **kwargs):
        self._file = resolve_path('~/potnanny/config.yml')
        self._yaml_loader = yaml.SafeLoader

        for k,v in kwargs.items():
            setattr(self, k, v)

        self._load_file()


    @classmethod
    async def load(cls):
        """
        Load config settings from database keychain.
        Ensure DB has been initiated before running this!
        """

        self = cls()
        try:
            obj = await Keychain.get(Keychain.name == 'settings')
            self._load_data(obj.attributes)
        except Exception as x:
            logger.warning(str(x))

        return self


    def __iter__(self):
        """
        Set up iter, so our object can be dumped correctly with dict(obj)
        """

        for k in self.__dict__.keys():
            if not k.startswith("_") and not k.startswith("load"):
                yield (k, self._ns_to_dict(getattr(self, k)))
            else:
                continue


    def _ns_to_dict(self, data:dict):
        """
        If item is a SimpleNamespace, convert to dict, else return item.

        args:
            an object
        returns:
            object, or dict
        """

        if type(data) is SimpleNamespace:
            return {k: self._ns_to_dict(v) for k, v in data.__dict__.items()}

        return data


    def _dict_to_ns(self, data:dict):
        """
        If item is a dict, convert to a SimpleNamespace, else return item.

        args:
            an object
        returns:
            object, or SimpleNamespace
        """

        if type(data) is not dict:
            return data

        ns = SimpleNamespace()
        for k,v in data.items():
            setattr(ns, k, self._dict_to_ns(v))

        return ns


    def _load_data(self, data:dict):
        """
        Set dict values as self attributes
        """

        for k, v in data.items():
            setattr(self, k, self._dict_to_ns(v))


    def _load_file(self):
        """
        Load config data from the default file
        """

        if not os.path.exists(self._file):
            logger.debug(f"Creating initial config file {self._file}")
            self._dump_yaml_config(DEFAULTS, self._file)

        logger.debug(f"Reading config {self._file}")
        with open(self._file, 'r', encoding='utf-8') as fh:
            data = yaml.load(fh, Loader=self._yaml_loader)
            self._load_data(data)


    def _dump(self):
        """
        Dump config data to default file
        """

        logger.debug(f"Dumping config to file {self._file}")
        self._dump_yaml_config(dict(self), self._file)


    def _dump_yaml_config(self, data:dict, path:str):
        """
        Dump dict as yaml to named file.

        args:
            - dict
            - pathname
        """

        with open(path, 'w', encoding='utf-8') as fh:
            yaml.dump(data, fh)



def convert_to_fahrenheit(celsius):
    """
    Convert a celsius temperature to fahrenheit.
    args:
        - a celsius number
    returns:
        - a fahrenheit number (float)
    """

    result = None
    try:
        result = (celsius * 1.8) + 32
    except:
        pass

    return result



import logging
from potnanny.database import db
from potnanny.plugins import BluetoothDevicePlugin
from potnanny.models.device import Device
from potnanny.utils.scanner import BLEScanner


logger = logging.getLogger(__name__)


class DiscoveryClient:
    def __init__(self, *args, **kwargs):
        self.scanner = BLEScanner()


    async def discover(self):
        results = []
        existing = await Device.select()
        found = await self.scanner.scan_devices()

        logger.debug(f"found devices: {found}")
        logger.debug(type(found))

        devices = self.recognized_devices([f for f in found])
        logger.debug(f"recognized devices: {devices}")
        for entry in devices:
            device, interface, attrs = entry
            if self.is_existing_device(existing, device, interface):
                continue

            opts = {
                'name': device.name,
                'interface': interface,
                'attributes': {'address': device.address},
                'room_id': None }

            # manufacturer bluetooth advertised names are often stupid, and
            # meaningless to the end user.
            # here, we replace it with the plugin name attribute to be more
            # helpful.
            try:
                klass = BluetoothDevicePlugin.get_named_class(opts['interface'])
                if klass:
                    opts['name'] = klass.name
            except:
                pass

            try:
                opts['attributes'].update(attrs)
            except:
                pass

            logger.debug(f"new device: {opts}")
            results.append(opts)

        if results:
            async with db.connection():
                for d in results:
                    print(f"new device {d}")
                    try:
                        obj = await Device.create(**d)
                        await obj.save()
                    except Exception as x:
                        logger.warning(x)

        return results


    def recognized_devices(self, devices:list) -> list:
        """
        Return a filtered list of ble devices that we can communicate with.
        [(device, interface_name), ]
        """

        results = []
        for d in devices:
            fingerprint = {'address': d.address, 'name': d.name}
            logger.debug(f" device: {d}, fingerprint: {fingerprint}")

            for p in BluetoothDevicePlugin.plugins:
                attrs = None

                if p.recognize_this(fingerprint):
                    found = '.'.join((p.__module__, p.__name__))
                    try:
                        attrs = p.attributes
                    except:
                        attrs = {}

                    results.append((d, found, attrs))

        return results


    def is_existing_device(self, existing:list, device, interface):
        """
        Compare ble-device to list of existing devices, and see if match
        exists already
        """

        def match_exists(e, plugin, address):
            try:
                if e.interface != plugin:
                    return False

                a1 = address.upper()
                a2 = e.attributes['address'].upper()
                if a1 == a2:
                    return True
            except Exception as x:
                logger.warning(str(x))

            return False

        for e in existing:
            if match_exists(e, interface, device.address) == True:
                return True

        return False


async def discover_new_devices():
    client = DiscoveryClient()
    await client.discover()



import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db, init_db
from potnanny.models.room import Room
from potnanny.models.device import Device


async def init_tables():
    await init_db(
        'aiosqlite:////tmp/db.db',
        pragmas=(('foreign_keys', 1),)
    )


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_foreignkey_cascade(self):
        async with db.connection():
            # create a parent
            room = await Room.create(name='room 1')
            await room.save()
            assert room.id > 0
            pk = room.id

            # create a child
            device = await Device.create(
                name='device 1', interface='foo.bar.baz', room_id=pk)
            await device.save()
            assert device.id > 0

            # delete the parent
            await room.delete_instance()

            # child should not exist
            results = await Device.select().where(Device.room_id == pk)
            assert len(results) == 0

if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
import datetime
from threading import Event
from potnanny.utils import utcnow
from potnanny.plugins.utils import load_plugins
from potnanny.models.device import Device
from potnanny.models.keychain import Keychain
from potnanny.controllers.collector import Collector
from potnanny.controllers.cleanup import purge_measurements


logger = logging.getLogger(__name__)
WORKER = None
WORKER_STOP = Event()
TASK = None


async def run_worker():
    """
    Convenience function, to run the worker
    """

    global TASK
    global WORKER
    WORKER_STOP.clear()

    logger.debug("Starting worker task")
    WORKER = Worker()
    TASK = asyncio.create_task(WORKER.run())


async def stop_worker():
    """
    Convenience function, to signal worker to stop
    """

    logger.debug("Stopping worker task")
    WORKER_STOP.set()


async def restart_worker():
    """
    Convenience function, to stop and restart the worker
    """

    await stop_worker()
    while TASK in asyncio.all_tasks():
        await asyncio.sleep(1)

    await run_worker()


class Worker:
    def __init__(self, *args, **kwargs):
        self.convert_c = False
        self.leaf_offset = -2
        self.polling_interval = 10
        self.plugin_path = "~/potnanny/plugins"
        self.storage_days = 7


    async def load_settings(self):
        results = await Keychain.select().where(Keychain.name == 'settings')
        obj = results[0]
        for k, v in obj.attributes.items():
            if hasattr(self, k):
                setattr(self, k, v)
            elif k == 'temperature_display':
                if v in ['f','F']:
                    self.convert_c = True

        load_plugins(self.plugin_path)


    async def run(self):
        await self.load_settings()

        while not WORKER_STOP.is_set():
            now = datetime.datetime.utcnow().replace(microsecond=0)

            # at midnight we clean old records
            if now.hour == 0 and now.minute == 0 and now.second == 0:
                start = now - datetime.timedelta(days=self.storage_days)
                logger.debug(f"cleaning up measurements older than {start}")
                t1 = asyncio.create_task(purge_measurements(start))

            if now.second == 0:
                # poll devices at intervals
                if now.minute % self.polling_interval == 0:
                    t3 = asyncio.create_task(self.poll_devices())

            await asyncio.sleep(1)


    async def poll_devices(self):
        """
        poll active devices for information
        """

        logger.debug("polling all devices")
        opts = {
            'convert_c': self.convert_c,
            'leaf_offset': self.leaf_offset}

        c = Collector(**opts)
        asyncio.create_task(c.collect())


    async def poll_device(self, pk:int):
        """
        poll device for information
        """

        logger.debug(f"polling single device ({pk})")
        opts = {
            'convert_c': self.convert_c,
            'leaf_offset': self.leaf_offset}

        c = Collector(**opts)
        asyncio.create_task(c.collect_id(pk))



import setuptools

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='potnanny',
    version='0.5.1',
    python_requires=">=3.11",
    packages=setuptools.find_packages(),
    description='Potnanny grow room automation controller.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='J Leary',
    author_email='potnanny@gmail.com',
    url='https://github.com/potnanny/application',
    install_requires=[
        'greenlet',
        'bleak',
        'aio-databases[aiosqlite]',
        'aiosmtplib',
        'aiofiles',
        'peewee-aio',
        'quart',
        'quart-auth',
        'quart-wtforms',
        'pyyaml',
        'gpiozero',
        'marshmallow',
        'pyyaml',
        'python-daemon',
        'markupsafe',
        'cryptography',
    ],
    package_data = {
        '': ['*.html', '*.css', '*.js', '*.woff2'],
    },
    entry_points = {
        'console_scripts': [
            'potnanny=potnanny.cli:main',
        ],
    },
)



import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.control import Control


async def init_tables():
    db.init('aiosqlite:////tmp/test.db')
    async with db.connection():
        await Control.create_table()


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        async with db.connection():
            control = await Control.create(name='control 1',
                outlet=1,
                attributes={'foo': 'bar'},
                device_id=1)
            await control.save()
            assert control.id > 0

    async def test_update(self):
        async with db.connection():
            control = await Control.create(name='control 2',
                outlet=1,
                attributes={'foo': 'bar!'},
                device_id=2)
            await control.save()

            control.attributes = {'foo': 'bar'}
            await control.save()
            assert control.attributes == {'foo': 'bar'}

    async def test_delete(self):
        async with db.connection():
            control = await Control.create(name='control 3',
                outlet=1,
                attributes={'foo': 'bar'},
                device_id=3)
            await control.save()
            pk = control.id
            await control.delete_instance()

            row = await Control.select().where(Control.name == 'control 3')
            assert len(row) == 0


if __name__ == '__main__':
    unittest.main()



import unittest
import peewee
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.keychain import Keychain


async def init_tables():
    db.init('aiosqlite:////tmp/test.db')
    async with db.connection():
        await Keychain.create_table()


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        async with db.connection():
            kc = await Keychain.create(name='settings',
                attributes={'temperature_display': 'F'})
            await kc.save()
            assert kc.id > 0

    async def test_update(self):
        async with db.connection():
            kc = await Keychain.create(name='foo',
                attributes={'bar': 13, 'baz': 42})
            await kc.save()

            kc.attributes.update({'baz': 43})
            await kc.save()
            assert kc.attributes == {'bar': 13, 'baz': 43}

    async def test_delete(self):
        async with db.connection():
            kc = await Keychain.create(name='bar',
                attributes={'one': 1})
            await kc.save()
            pk = kc.id
            await kc.delete_instance()

            row = await Keychain.select().where(Keychain.id == pk)
            assert len(row) == 0

    async def test_unique_name(self):
        with self.assertRaises(peewee.IntegrityError):
            async with db.connection():
                kc1 = await Keychain.create(name='user',
                    attributes={'name': 'foo'})
                await kc1.save()

                kc2 = await Keychain.create(name='user',
                    attributes={'name': 'foo'})
                await kc2.save()


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.measurement import Measurement


async def init_tables():
    db.init('aiosqlite:////tmp/test.db')
    async with db.connection():
        await Measurement.create_table()


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        async with db.connection():
            meas = await Measurement.create(
                type='temperature', value=68.0, device_id=1)
            await meas.save()
            assert meas.id > 0

    async def test_update(self):
        async with db.connection():
            meas = await Measurement.create(
                type='humidity', value=55.0, device_id=1)
            await meas.save()

            meas.value = 55
            await meas.save()
            assert meas.value == 55.0

    async def test_delete(self):
        async with db.connection():
            meas = await Measurement.create(
                type='vpd', value=1.0, device_id=1)
            await meas.save()
            pk = meas.id
            await meas.delete_instance()

            row = await Measurement.select().where(Measurement.id == pk)
            assert len(row) == 0


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.user import User
from potnanny.utils.password import verify_password, hash_password


async def init_tables():
    db.init('aiosqlite:////tmp/test.db')
    async with db.connection():
        await User.create_table()


class TestUtils(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        pass

    async def asyncTearDown(self):
        pass

    async def test_hashing(self):
        pw = hash_password('foobar!')
        print(pw)
        assert pw

    async def test_verify(self):
        pw = hash_password('foobar!')
        assert verify_password('foobar!', pw) is True
        assert verify_password('foobar!', 'foobar') is False

if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.database import db
from potnanny.models.room import Room, RoomSchema


async def init_tables():
    db.init('aiosqlite:////tmp/test.db')
    async with db.connection():
        await Room.create_table()


class TestSchemas(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        pass

    async def asyncTearDown(self):
        pass

    async def test_basic_schema(self):
        jsondata = {'name': 'test room', 'foo': 'bar'}
        schema = RoomSchema()
        data = schema.load(jsondata)
        assert data is not None
        assert 'foo' not in data


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        await init_tables()

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        async with db.connection():
            room = await Room.create(name='room 1')
            await room.save()
            assert room.id > 0

        assert isinstance(room.as_dict(), dict)


    async def test_update(self):
        async with db.connection():
            room = await Room.create(name='room 2')
            await room.save()

            room.notes = 'test notes'
            await room.save()
            assert room.notes == 'test notes'

        assert isinstance(room.as_dict(), dict)


    async def test_delete(self):
        async with db.connection():
            room = await Room.create(name='room 3')
            await room.save()
            pk = room.id
            await room.delete_instance()

            row = await Room.select().where(Room.name == 'room 3')
            assert len(row) == 0


if __name__ == '__main__':
    unittest.main()



import os
import base64
import string
import secrets
from werkzeug.security import generate_password_hash, check_password_hash


def hash_password(password:str) -> str:
    secret = generate_password_hash(password)
    return secret


def verify_password(guessed:str, hashed:str) -> bool:
    return check_password_hash(hashed, guessed)


def random_key(length:int = 24) -> str:
    choices = string.ascii_lowercase + string.ascii_uppercase + string.digits
    symbols = list(choices)
    password = ""
    while len(password) < length:
        password += secrets.choice(symbols)

    return password



import logging
import asyncio
import datetime
import marshmallow
import potnanny.controllers.worker as worker
from peewee_aio import fields
from potnanny.database import BaseModel
from potnanny.controllers.outlet import switch_device_outlet
from potnanny.models.schemas.safe import SafeSchema
from potnanny.utils.eval import evaluate
from .device import Device
from .ext import JSONField


logger = logging.getLogger(__name__)


class ControlSchema(SafeSchema):
    name = marshmallow.fields.String()
    device_id = marshmallow.fields.Integer()
    outlet = marshmallow.fields.Integer()
    attributes = marshmallow.fields.Dict(allow_none=True)


class Control(BaseModel):
    id = fields.AutoField()
    name = fields.CharField(48)
    outlet = fields.IntegerField()
    attributes = JSONField(default={})
    created = fields.DateTimeField(default=datetime.datetime.utcnow)
    device = fields.ForeignKeyField(Device,
        on_delete='CASCADE',
        backref='controls' )


    def __str__(self):
        return f"<Control id={self.id}, name='{self.name}'>"


    def as_dict(self) -> dict:
        return {
            'id': self.id,
            'name': self.name,
            'outlet': self.outlet,
            'device_id': self.device_id,
            'attributes': self.attributes,
            'created': self.created.isoformat() + "Z",
        }


    async def input(self, data:dict):
        """
        Accept measurement data as input, evaluate, and perform on/off control
        as required:
        """

        if self._precheck(data) is not True:
            return

        states = {'on': 1, 'off': 0}
        for key in ('on', 'off'):
            if (key not in self.attributes or not self.attributes[key] or
                self.attributes[key]['condition'] in [None, "", "null"] or
                self.attributes[key]['threshold'] in [None, "", "null"]):
                continue

            working = self.attributes[key]
            equation = "%0.1f %s %s" % (
                float(data['value']),
                working['condition'],
                working['threshold'])

            try:
                if evaluate(equation) is True:
                    state = states[key]

                    # switch device to state
                    t1 = asyncio.create_task(self.switch_task(
                        self.device_id,
                        self.outlet,
                        state))

                    # timed switch? set back to original state after N seconds...
                    if 'seconds' in working and int(working['seconds']) > 0:
                        t2 = asyncio.create_task(self.switch_task(
                            self.device_id,
                            self.outlet,
                            int(not state),
                            int(working['seconds'])
                        ))
            except Exception as x:
                logger.warning(x)


    def _precheck(self, data:dict) -> bool:
        """
        examine incoming data, decide if it is valid for us
        """

        try:
            if data['device_id'] != self.attributes['input_device_id']:
                return False
        except Exception as x:
            logger.warning(str(x))
            return False

        try:
            if data['type'] != self.attributes['type']:
                return False
        except Exception as x:
            logger.warning(str(x))
            return False

        return True


    async def switch_task(self, device:int, outlet:int, state:int, delay:int = 0):
        try:
            if delay:
                await asyncio.sleep(delay)
                rval = await switch_device_outlet(device, outlet, state)
                return rval
            else:
                rval = await switch_device_outlet(device, outlet, state)
                return rval
        except Exception as x:
            logger.warning(x)

        return -1



def is_number(val):
    """
    check if a string can be cast to an int or float successfully.
    args:
        - str ("13.0", "42")
    returns:
        - type of number it matches (int, float) or None
    """

    if type(val) is str and val.isdigit():
        return int

    try:
        f = float(val)
        return float
    except:
        pass

    return None



import re
import logging


logger = logging.getLogger(__name__)


class FingerprintMixin:
    """
    This is a class mixin for BLE plugins, to be able to recognize a
    BLE device that can be communicated with (by its device address or name)

    Each entry in a class fingerprint definition must match, in order for
    this to return True.
    """

    @classmethod
    def recognize_this(cls, fp):
        if not hasattr(cls, 'fingerprint'):
            logger.warning("Class %s no fingerprint defined" % cls)
            return False

        for key, regex in cls.fingerprint.items():
            if key not in fp:
                return False

            if not regex.search(fp[key]):
                return False

        return True



from quart import Blueprint, render_template, abort
from jinja2 import TemplateNotFound


bp = Blueprint('about', __name__, template_folder='templates')


@bp.route('/about', methods=['GET'])
async def index():
    try:
        return await render_template('about/index.html')
    except TemplateNotFound:
        abort(404)



import unittest
import datetime
import calendar
from unittest import IsolatedAsyncioTestCase
from potnanny.models.weekday import WeekdayMap


class TestModels(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        pass

    async def asyncTearDown(self):
        pass

    async def test_lookup(self):
        now:datetime.datetime = datetime.datetime.now()
        dow = calendar.day_name[now.weekday()]


if __name__ == '__main__':
    unittest.main()



import logging
import datetime
import copy
from peewee import fn
from potnanny.utils import iso_from_sqlite
from potnanny.models.room import Room
from potnanny.models.device import Device
from potnanny.models.measurement import Measurement
from potnanny.models.keychain import Keychain
from potnanny.database import db


logger = logging.getLogger(__name__)


DEFAULT_GRAPH = {
    'type': 'line',
    'options': {
        'maintainAspectRatio': False,
        'responsive': True,
        'plugins': {
            'title': {
                'display': True,
                'text': None,
            }
        },
    },
    'data': {
        'datasets': []
    }
}


async def room_graph(pk:int, mtype:str, hours:int = 12):
    """
    Get graph data for room
    """

    try:
        kc = await Keychain.select().where(Keychain.name == 'settings').first()
        hours = kc.attributes['graph_hours']
    except:
        pass

    now = datetime.datetime.utcnow()
    start = str(now - datetime.timedelta(hours=hours))
    graph = copy.deepcopy(DEFAULT_GRAPH)

    # one label tick per hour only
    # graph['options']['scales']['xAxes']['ticks']['maxTicksLimit'] = hours

    sql = """
        SELECT
          d.id, d.name, m.type, m.value, m.created
        FROM
          measurement as m,
          device as d
        WHERE
            d.room_id = %d
        AND
            m.device_id = d.id
        AND
            m.created >= '%s'
    """ % (pk, start)

    if mtype is not None:
        sql += """
        AND
            m.type = '%s' """ % mtype

    sql += """
        ORDER BY
            m.created"""

    data = {}
    labels = []
    async with db.connection():
        rows = await db.fetchall(sql)

    # first, built labels
    for r in rows:
        try:
            dt = iso_from_sqlite(r[4])
            labels.append(dt)
        except:
            pass

    labels = sorted(list(set(labels)))
    graph['data'].update({'labels': labels})

    for r in rows:
        if graph['options']['plugins']['title']['text'] is None:
            graph['options']['plugins']['title']['text'] = mtype.upper()

        if r[0] not in data:
            data[r[0]] = {
                'label': r[1],
                'data': [{'y': None, 'x': t} for t in labels] }
        dt = iso_from_sqlite(r[4])
        for item in data[r[0]]['data']:
            if item['x'] == dt:
                item['y'] = r[3]
                break

    # graph['data']['labels'] = sorted(list(set(labels)))
    graph['data']['labels'] = sorted(labels)
    graph['data']['datasets'] = [v for v in data.values()]

    return graph


async def device_graph(pk:int, mtype:str, hours:int = 12):
    """
    Get graph data for device
    """

    try:
        kc = await Keychain.select().where(Keychain.name == 'settings').first()
        hours = kc.attributes['graph_hours']
    except:
        pass

    now = datetime.datetime.utcnow()
    start = str(now - datetime.timedelta(hours=hours))
    graph = copy.deepcopy(DEFAULT_GRAPH)

    # one label tick per hour only
    # graph['options']['scales']['xAxes']['ticks']['maxTicksLimit'] = hours

    sql = """
        SELECT
          d.id, d.name, m.type, m.value, m.created
        FROM
          measurement as m,
          device as d
        WHERE
            d.id = %d
        AND
            m.device_id = d.id
        AND
            m.created >= '%s'
    """ % (pk, start)

    if mtype is not None:
        sql += """
        AND
            m.type = '%s' """ % mtype

    sql += """
        ORDER BY
            m.created """

    data = {}
    labels = []
    async with db.connection():
        rows = await db.fetchall(sql)

    for r in rows:
        if graph['options']['plugins']['title']['text'] is None:
            graph['options']['plugins']['title']['text'] = r[1]

        if r[2] not in data:
            data[r[2]] = {
                'label': r[2],
                'data': [] }
        dt = iso_from_sqlite(r[4])
        if dt not in labels:
            labels.append(dt)

        data[r[2]]['data'].append({
            'y': r[3],
            'x': dt,
        })

    graph['data']['labels'] = sorted(list(set(labels)))
    graph['data']['datasets'] = [v for v in data.values()]
    return graph
